from enum import Enum
from typing import Any, Literal

import pydantic
from pydantic import (
    BaseModel,
    ConfigDict,
    SerializationInfo,
    field_validator,
    model_serializer,
)
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import override

from kpops.components.base_components.helm_app import HelmAppValues
from kpops.components.base_components.models.topic import KafkaTopic, KafkaTopicStr
from kpops.utils.pydantic import (
    CamelCaseConfigModel,
    DescConfigModel,
    by_alias,
    exclude_by_value,
    to_dot,
)


class KafkaConnectorType(str, Enum):
    SINK = "sink"
    SOURCE = "source"


class KafkaConnectorConfig(DescConfigModel):
    """Settings specific to Kafka Connectors."""

    connector_class: str
    name: SkipJsonSchema[str]
    topics: list[KafkaTopicStr] = []
    topics_regex: str | None = None
    errors_deadletterqueue_topic_name: KafkaTopicStr | None = None

    @override
    @staticmethod
    def json_schema_extra(schema: dict[str, Any], model: type[BaseModel]) -> None:
        super(KafkaConnectorConfig, KafkaConnectorConfig).json_schema_extra(
            schema, model
        )
        schema["additional_properties"] = {"type": "string"}

    model_config = ConfigDict(
        extra="allow",
        alias_generator=to_dot,
        json_schema_extra=json_schema_extra,
        populate_by_name=True,
    )

    @field_validator("connector_class")
    @classmethod
    def connector_class_must_contain_dot(cls, connector_class: str) -> str:
        if "." not in connector_class:
            msg = f"Invalid connector class {connector_class}"
            raise ValueError(msg)
        return connector_class

    @pydantic.field_validator("topics", mode="before")
    @classmethod
    def deserialize_topics(cls, topics: Any) -> list[KafkaTopic] | None | Any:
        if isinstance(topics, str):
            return [KafkaTopic(name=topic_name) for topic_name in topics.split(",")]
        return topics

    @property
    def class_name(self) -> str:
        return self.connector_class.split(".")[-1]

    @pydantic.field_serializer("topics")
    def serialize_topics(self, topics: list[KafkaTopic]) -> str | None:
        if not topics:
            return None
        return ",".join(topic.name for topic in topics)

    # TODO(Ivan Yordanov): Currently hacky and potentially unsafe. Find cleaner solution
    @model_serializer(mode="wrap", when_used="always")
    def serialize_model(self, handler, info: SerializationInfo) -> dict[str, Any]:
        result = exclude_by_value(handler(self), None)
        return {by_alias(self, name): value for name, value in result.items()}


class ConnectorTask(BaseModel):
    connector: str
    task: int


class KafkaConnectResponse(BaseModel):
    name: str
    config: dict[str, str]
    tasks: list[ConnectorTask]
    type: str | None = None

    model_config = ConfigDict(extra="forbid")


class KafkaConnectConfigError(BaseModel):
    name: str
    errors: list[str]


class KafkaConnectConfigDescription(BaseModel):
    value: KafkaConnectConfigError


class KafkaConnectConfigErrorResponse(BaseModel):
    name: str
    error_count: int
    configs: list[KafkaConnectConfigDescription]


class KafkaConnectorResetterConfig(CamelCaseConfigModel):
    brokers: str
    connector: str
    delete_consumer_group: bool | None = None
    offset_topic: str | None = None


class KafkaConnectorResetterValues(HelmAppValues):
    connector_type: Literal["source", "sink"]
    config: KafkaConnectorResetterConfig



from __future__ import annotations

import logging

from kpops.component_handlers.topic.exception import (
    TopicNotFoundException,
    TopicTransactionError,
)
from kpops.component_handlers.topic.model import (
    TopicConfigResponse,
    TopicResponse,
    TopicSpec,
)
from kpops.component_handlers.topic.proxy_wrapper import HEADERS, ProxyWrapper
from kpops.component_handlers.topic.utils import (
    get_effective_config,
    parse_and_compare_topic_configs,
    parse_rest_proxy_topic_config,
)
from kpops.components.base_components.models.topic import KafkaTopic
from kpops.utils.colorify import greenify, magentaify
from kpops.utils.dict_differ import Diff, DiffType, render_diff

log = logging.getLogger("KafkaTopic")


class TopicHandler:
    def __init__(self, proxy_wrapper: ProxyWrapper) -> None:
        self.proxy_wrapper = proxy_wrapper

    async def create_topic(self, topic: KafkaTopic, dry_run: bool) -> None:
        """Create a new Kafka topic or update topic configuration if it already exists.

        :param topic: Kafka topic to be created or updated
        :param dry_run: Whether to do a dry run without making changes
        :raises TopicTransactionError: Partition count of topic changed
        :raises TopicTransactionError: Replication factor of topic changed
        """
        topic_spec = self.__prepare_body(topic)
        if dry_run:
            await self.__dry_run_topic_creation(topic, topic_spec)
        else:
            await self.__execute_topic_creation(topic, topic_spec)

    async def delete_topic(self, topic: KafkaTopic, dry_run: bool) -> None:
        """Delete an existing Kafka topic.

        :param topic: Kafka topic to be deleted
        :param dry_run: Whether to do a dry run without making changes
        """
        if dry_run:
            await self.__dry_run_topic_deletion(topic.name)
        else:
            await self.__execute_topic_deletion(topic.name)

    @staticmethod
    def __get_topic_config_diff(
        cluster_config: TopicConfigResponse, current_config: dict
    ) -> list[Diff]:
        comparable_in_cluster_config_dict, _ = parse_rest_proxy_topic_config(
            cluster_config
        )
        return list(Diff.from_dicts(comparable_in_cluster_config_dict, current_config))

    async def __dry_run_topic_creation(
        self, topic: KafkaTopic, topic_spec: TopicSpec
    ) -> None:
        try:
            topic_in_cluster = await self.proxy_wrapper.get_topic(topic.name)
            topic_name = topic_in_cluster.topic_name
            topic_config_in_cluster = await self.proxy_wrapper.get_topic_config(
                topic_name
            )
            in_cluster_config, new_config = parse_and_compare_topic_configs(
                topic_config_in_cluster, topic.config.configs
            )
            if diff := render_diff(in_cluster_config, new_config):
                log.info(f"Config changes for topic {topic_name}:")
                log.info("\n" + diff)

            log.info(f"Topic Creation: {topic_name} already exists in cluster.")
            log.debug("HTTP/1.1 400 Bad Request")
            log.debug(HEADERS)
            error_message = {
                "error_code": 40002,
                "message": f"Topic '{topic_name}' already exists.",
            }
            log.debug(error_message)

            broker_config = await self.proxy_wrapper.get_broker_config()
            effective_config = get_effective_config(broker_config)

            self.__check_partition_count(topic_in_cluster, topic_spec, effective_config)
            self.__check_replication_factor(
                topic_in_cluster, topic_spec, effective_config
            )
        except TopicNotFoundException:
            log.info(
                greenify(
                    f"Topic Creation: {topic.name} does not exist in the cluster. Creating topic."
                )
            )
            log.debug(f"POST /clusters/{self.proxy_wrapper.cluster_id}/topics HTTP/1.1")
            log.debug(f"Host: {self.proxy_wrapper.url}")
            log.debug(HEADERS)
            log.debug(topic_spec.model_dump())

    async def __execute_topic_creation(
        self, topic: KafkaTopic, topic_spec: TopicSpec
    ) -> None:
        try:
            await self.proxy_wrapper.get_topic(topic.name)
            topic_config_in_cluster = await self.proxy_wrapper.get_topic_config(
                topic.name
            )
            differences = self.__get_topic_config_diff(
                topic_config_in_cluster, topic.config.configs
            )

            if differences:
                json_body = []
                for difference in differences:
                    if difference.diff_type is DiffType.REMOVE:
                        json_body.append(
                            {"name": difference.key, "operation": "DELETE"}
                        )
                    elif config_value := difference.change.new_value:
                        json_body.append(
                            {"name": difference.key, "value": config_value}
                        )
                await self.proxy_wrapper.batch_alter_topic_config(topic.name, json_body)

            else:
                log.info(
                    f"Topic Creation: config of topic {topic.name} didn't change. Skipping update."
                )
        except TopicNotFoundException:
            await self.proxy_wrapper.create_topic(topic_spec)

    @staticmethod
    def __check_partition_count(
        topic_in_cluster: TopicResponse,
        topic_spec: TopicSpec,
        broker_config: dict[str, str],
    ) -> None:
        topic_name = topic_in_cluster.topic_name
        partition_count = topic_in_cluster.partitions_count
        if partition_count == (
            topic_spec.partitions_count or int(broker_config["num.partitions"])
        ):
            log.debug(
                f"Topic Creation: partition count of topic {topic_name} did not change. Current partitions count {partition_count}. Updating configs."
            )
        else:
            msg = f"Topic Creation: partition count of topic {topic_name} changed! Partitions count of topic {topic_name} is {partition_count}. The given partitions count {topic_spec.partitions_count}."
            raise TopicTransactionError(msg)

    @staticmethod
    def __check_replication_factor(
        topic_in_cluster: TopicResponse,
        topic_spec: TopicSpec,
        broker_config: dict[str, str],
    ) -> None:
        topic_name = topic_in_cluster.topic_name
        replication_factor = topic_in_cluster.replication_factor
        if replication_factor == (
            topic_spec.replication_factor
            or int(broker_config["default.replication.factor"])
        ):
            log.debug(
                f"Topic Creation: replication factor of topic {topic_name} did not change. Current replication factor {replication_factor}. Updating configs."
            )
        else:
            msg = f"Topic Creation: replication factor of topic {topic_name} changed! Replication factor of topic {topic_name} is {replication_factor}. The given replication count {topic_spec.replication_factor}."
            raise TopicTransactionError(msg)

    async def __dry_run_topic_deletion(self, topic_name: str) -> None:
        try:
            topic_in_cluster = await self.proxy_wrapper.get_topic(topic_name)
            log.info(
                magentaify(
                    f"Topic Deletion: topic {topic_in_cluster.topic_name} exists in the cluster. Deleting topic."
                )
            )
            log.debug(
                f"DELETE /clusters/{self.proxy_wrapper.cluster_id}/topics HTTP/1.1"
            )
        except TopicNotFoundException:
            log.warning(
                f"Topic Deletion: topic {topic_name} does not exist in the cluster and cannot be deleted. Skipping."
            )
            log.debug(f"Host: {self.proxy_wrapper.url}")
            log.debug(HEADERS)
            log.debug("HTTP/1.1 404 Not Found")
            log.debug(HEADERS)
            error_message = {
                "error_code": 40403,
                "message": f"This server does not host the topic-partition '{topic_name}'.",
            }
            log.debug(error_message)

    async def __execute_topic_deletion(self, topic_name: str) -> None:
        try:
            await self.proxy_wrapper.get_topic(topic_name)
            await self.proxy_wrapper.delete_topic(topic_name)
        except TopicNotFoundException:
            log.warning(
                f"Topic Deletion: topic {topic_name} does not exist in the cluster and cannot be deleted. Skipping."
            )

    @classmethod
    def __prepare_body(cls, topic: KafkaTopic) -> TopicSpec:
        """Prepare the POST request body needed for the topic creation.

        :param topic_name: The name of the topic
        :param topic_config: The topic config
        :return: Topic specification
        """
        topic_spec_json: dict = topic.config.model_dump(
            include={
                "partitions_count": True,
                "replication_factor": True,
                "configs": True,
            },
            exclude_none=True,
        )
        configs = []
        for config_name, config_value in topic_spec_json["configs"].items():
            configs.append({"name": config_name, "value": config_value})
        topic_spec_json["configs"] = configs
        topic_spec_json["topic_name"] = topic.name
        return TopicSpec(**topic_spec_json)



from enum import Enum
from typing import Any, NewType

from pydantic import ConfigDict, Field, model_validator

from kpops.components.base_components.models import TopicName
from kpops.utils.docstring import describe_attr
from kpops.utils.pydantic import DescConfigModel


class InputTopicTypes(str, Enum):
    """Input topic types.

    - INPUT: input topic
    - PATTERN: extra-topic-pattern or input-topic-pattern
    """

    INPUT = "input"
    PATTERN = "pattern"


class FromTopic(DescConfigModel):
    """Input topic.

    :param type: Topic type, defaults to None
    :param role: Custom identifier belonging to a topic;
        define only if `type` is `pattern` or `None`, defaults to None
    """

    type: InputTopicTypes | None = Field(
        default=None, description=describe_attr("type", __doc__)
    )
    role: str | None = Field(default=None, description=describe_attr("role", __doc__))

    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
    )

    @model_validator(mode="after")
    def extra_topic_role(self) -> Any:
        """Ensure that `cls.role` is used correctly, assign type if needed."""
        if self.type == InputTopicTypes.INPUT and self.role:
            msg = "Define role only if `type` is `pattern` or `None`"
            raise ValueError(msg)
        return self


ComponentName = NewType("ComponentName", str)


class FromSection(DescConfigModel):
    """Holds multiple input topics.

    :param topics: Input topics
    :param components: Components to read from
    """

    topics: dict[TopicName, FromTopic] = Field(
        default={},
        description=describe_attr("topics", __doc__),
    )
    components: dict[ComponentName, FromTopic] = Field(
        default={},
        description=describe_attr("components", __doc__),
    )

    model_config = ConfigDict(
        extra="forbid",
    )



from pathlib import Path
from typing import Any

import yaml
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from kpops.config import KpopsConfig
from kpops.utils.docstring import describe_object
from kpops.utils.json import is_jsonable
from kpops.utils.pydantic import collect_fields, issubclass_patched


def extract_config_fields_for_yaml(
    fields: dict[str, Any], required: bool
) -> dict[str, Any]:
    """Return only (non-)required fields and their respective default values.

    :param fields: Dict containing the fields to be categorized. The key of a
        record is the name of the field, the value is the field's type.
    :param required: Whether to extract only the required fields or only the
        non-required ones.
    """
    extracted_fields = {}
    for key, value in fields.items():
        if issubclass(type(value), FieldInfo):
            if required and value.default in [PydanticUndefined, Ellipsis]:
                extracted_fields[key] = None
            elif not (required or value.default in [PydanticUndefined, Ellipsis]):
                if is_jsonable(value.default):
                    extracted_fields[key] = value.default
                elif issubclass_patched(value.default, BaseModel):
                    extracted_fields[key] = value.default.model_dump(mode="json")
                else:
                    extracted_fields[key] = str(value.default)
        else:
            extracted_fields[key] = extract_config_fields_for_yaml(
                fields[key], required
            )
    return extracted_fields


def create_config(file_name: str, dir_path: Path, include_optional: bool) -> None:
    """Create a KPOps config yaml.

    :param file_name: Name for the file
    :param dir_path: Directory in which the file should be created
    :param include_optional: Whether to include non-required settings
    """
    file_path = Path(dir_path / (file_name + ".yaml"))
    file_path.touch(exist_ok=False)
    with file_path.open(mode="w") as conf:
        conf.write("# " + describe_object(KpopsConfig.__doc__))  # Write title
        non_required = extract_config_fields_for_yaml(
            collect_fields(KpopsConfig), False
        )
        required = extract_config_fields_for_yaml(collect_fields(KpopsConfig), True)
        for k in non_required:
            required.pop(k, None)
        conf.write("\n\n# Required fields\n")
        conf.write(yaml.dump(required))
        if include_optional:
            conf.write("\n# Non-required fields\n")
            conf.write(yaml.dump(non_required))


def init_project(path: Path, conf_incl_opt: bool):
    """Initiate a default empty project.

    :param path: Directory in which the project should be initiated
    :param conf_incl_opt: Whether to include non-required settings
        in the generated config file
    """
    create_config("config", path, conf_incl_opt)
    Path(path / "pipeline.yaml").touch(exist_ok=False)
    Path(path / "defaults.yaml").touch(exist_ok=False)



from __future__ import annotations

from typing import Any

import pydantic
from pydantic import BaseModel, ConfigDict, Field, model_validator

from kpops.components.base_components.kafka_app import (
    KafkaAppValues,
    KafkaStreamsConfig,
)
from kpops.components.base_components.models.topic import KafkaTopic, KafkaTopicStr
from kpops.pipeline import ValidationError
from kpops.utils.docstring import describe_attr
from kpops.utils.pydantic import (
    CamelCaseConfigModel,
    DescConfigModel,
)


class StreamsConfig(KafkaStreamsConfig):
    """Streams Bootstrap streams section.

    :param input_topics: Input topics, defaults to []
    :param input_pattern: Input pattern, defaults to None
    :param extra_input_topics: Extra input topics, defaults to {}
    :param extra_input_patterns: Extra input patterns, defaults to {}
    :param error_topic: Error topic, defaults to None
    :param config: Configuration, defaults to {}
    :param delete_output: Whether the output topics with their associated schemas and the consumer group should be deleted during the cleanup, defaults to None
    """

    input_topics: list[KafkaTopicStr] = Field(
        default=[], description=describe_attr("input_topics", __doc__)
    )
    input_pattern: str | None = Field(
        default=None, description=describe_attr("input_pattern", __doc__)
    )
    extra_input_topics: dict[str, list[KafkaTopicStr]] = Field(
        default={}, description=describe_attr("extra_input_topics", __doc__)
    )
    extra_input_patterns: dict[str, str] = Field(
        default={}, description=describe_attr("extra_input_patterns", __doc__)
    )
    error_topic: KafkaTopicStr | None = Field(
        default=None, description=describe_attr("error_topic", __doc__)
    )
    config: dict[str, Any] = Field(
        default={}, description=describe_attr("config", __doc__)
    )
    delete_output: bool | None = Field(
        default=None, description=describe_attr("delete_output", __doc__)
    )

    @pydantic.field_validator("input_topics", mode="before")
    @classmethod
    def deserialize_input_topics(cls, input_topics: Any) -> list[KafkaTopic] | Any:
        if isinstance(input_topics, list):
            return [KafkaTopic(name=topic_name) for topic_name in input_topics]
        return input_topics

    @pydantic.field_validator("extra_input_topics", mode="before")
    @classmethod
    def deserialize_extra_input_topics(
        cls, extra_input_topics: Any
    ) -> dict[str, list[KafkaTopic]] | Any:
        if isinstance(extra_input_topics, dict):
            return {
                role: [KafkaTopic(name=topic_name) for topic_name in topics]
                for role, topics in extra_input_topics.items()
            }
        return extra_input_topics

    @pydantic.field_serializer("input_topics")
    def serialize_topics(self, topics: list[KafkaTopic]) -> list[str]:
        return [topic.name for topic in topics]

    @pydantic.field_serializer("extra_input_topics")
    def serialize_extra_input_topics(
        self, extra_topics: dict[str, list[KafkaTopic]]
    ) -> dict[str, list[str]]:
        return {
            role: self.serialize_topics(topics) for role, topics in extra_topics.items()
        }

    def add_input_topics(self, topics: list[KafkaTopic]) -> None:
        """Add given topics to the list of input topics.

        Ensures no duplicate topics in the list.

        :param topics: Input topics
        """
        self.input_topics = KafkaTopic.deduplicate(self.input_topics + topics)

    def add_extra_input_topics(self, role: str, topics: list[KafkaTopic]) -> None:
        """Add given extra topics that share a role to the list of extra input topics.

        Ensures no duplicate topics in the list.

        :param topics: Extra input topics
        :param role: Topic role
        """
        self.extra_input_topics[role] = KafkaTopic.deduplicate(
            self.extra_input_topics.get(role, []) + topics
        )


class StreamsAppAutoScaling(CamelCaseConfigModel, DescConfigModel):
    """Kubernetes Event-driven Autoscaling config.

    :param enabled: Whether to enable auto-scaling using KEDA., defaults to False
    :param consumer_group: Name of the consumer group used for checking the
        offset on the topic and processing the related lag.
        Mandatory to set when auto-scaling is enabled.
    :param lag_threshold: Average target value to trigger scaling actions.
        Mandatory to set when auto-scaling is enabled.
    :param polling_interval: This is the interval to check each trigger on.
        https://keda.sh/docs/2.9/concepts/scaling-deployments/#pollinginterval,
        defaults to 30
    :param cooldown_period: The period to wait after the last trigger reported
        active before scaling the resource back to 0.
        https://keda.sh/docs/2.9/concepts/scaling-deployments/#cooldownperiod,
        defaults to 300
    :param offset_reset_policy: The offset reset policy for the consumer if the
        consumer group is not yet subscribed to a partition.,
        defaults to "earliest"
    :param min_replicas: Minimum number of replicas KEDA will scale the resource down to.
        "https://keda.sh/docs/2.9/concepts/scaling-deployments/#minreplicacount",
        defaults to 0
    :param max_replicas: This setting is passed to the HPA definition that KEDA
        will create for a given resource and holds the maximum number of replicas
        of the target resouce.
        https://keda.sh/docs/2.9/concepts/scaling-deployments/#maxreplicacount,
        defaults to 1
    :param idle_replicas: If this property is set, KEDA will scale the resource
        down to this number of replicas.
        https://keda.sh/docs/2.9/concepts/scaling-deployments/#idlereplicacount,
        defaults to None
    :param topics: List of auto-generated Kafka Streams topics used by the streams app.,
        defaults to []
    """

    enabled: bool = Field(
        default=False,
        description=describe_attr("streams", __doc__),
    )
    consumer_group: str | None = Field(
        default=None,
        title="Consumer group",
        description=describe_attr("consumer_group", __doc__),
    )
    lag_threshold: int | None = Field(
        default=None,
        title="Lag threshold",
        description=describe_attr("lag_threshold", __doc__),
    )
    polling_interval: int = Field(
        default=30,
        title="Polling interval",
        description=describe_attr("polling_interval", __doc__),
    )
    cooldown_period: int = Field(
        default=300,
        title="Cooldown period",
        description=describe_attr("cooldown_period", __doc__),
    )
    offset_reset_policy: str = Field(
        default="earliest",
        title="Offset reset policy",
        description=describe_attr("offset_reset_policy", __doc__),
    )
    min_replicas: int = Field(
        default=0,
        title="Min replica count",
        description=describe_attr("min_replicas", __doc__),
    )
    max_replicas: int = Field(
        default=1,
        title="Max replica count",
        description=describe_attr("max_replicas", __doc__),
    )
    idle_replicas: int | None = Field(
        default=None,
        title="Idle replica count",
        description=describe_attr("idle_replicas", __doc__),
    )
    topics: list[str] = Field(
        default=[],
        description=describe_attr("topics", __doc__),
    )
    model_config = ConfigDict(extra="allow")

    @model_validator(mode="after")
    def validate_mandatory_fields_are_set(
        self: StreamsAppAutoScaling,
    ) -> StreamsAppAutoScaling:  # TODO: typing.Self for Python 3.11+
        if self.enabled and (self.consumer_group is None or self.lag_threshold is None):
            msg = (
                "If app.autoscaling.enabled is set to true, "
                "the fields app.autoscaling.consumer_group and app.autoscaling.lag_threshold should be set."
            )
            raise ValidationError(msg)
        return self


class PersistenceConfig(BaseModel):
    """streams-bootstrap persistence configurations.

    :param enabled: Whether to use a persistent volume to store the state of the streams app.
    :param size: The size of the PersistentVolume to allocate to each streams pod in the StatefulSet.
    :param storage_class: Storage class to use for the persistent volume.
    """

    enabled: bool = Field(
        default=False,
        description="Whether to use a persistent volume to store the state of the streams app.	",
    )
    size: str | None = Field(
        default=None,
        description="The size of the PersistentVolume to allocate to each streams pod in the StatefulSet.",
    )
    storage_class: str | None = Field(
        default=None,
        description="Storage class to use for the persistent volume.",
    )

    @model_validator(mode="after")
    def validate_mandatory_fields_are_set(
        self: PersistenceConfig,
    ) -> PersistenceConfig:  # TODO: typing.Self for Python 3.11+
        if self.enabled and self.size is None:
            msg = (
                "If app.persistence.enabled is set to true, "
                "the field app.persistence.size needs to be set."
            )
            raise ValidationError(msg)
        return self


class StreamsAppValues(KafkaAppValues):
    """streams-bootstrap app configurations.

    The attributes correspond to keys and values that are used as values for the streams bootstrap helm chart.

    :param streams: streams-bootstrap streams section
    :param autoscaling: Kubernetes event-driven autoscaling config, defaults to None
    """

    streams: StreamsConfig = Field(
        default=...,
        description=describe_attr("streams", __doc__),
    )
    autoscaling: StreamsAppAutoScaling | None = Field(
        default=None,
        description=describe_attr("autoscaling", __doc__),
    )
    stateful_set: bool = Field(
        default=False,
        description="Whether to use a Statefulset instead of a Deployment to deploy the streams app.",
    )
    persistence: PersistenceConfig = Field(
        default=PersistenceConfig(),
        description=describe_attr("persistence", __doc__),
    )
    model_config = ConfigDict(extra="allow")



from __future__ import annotations

from collections.abc import Iterable
from enum import Enum
from typing import Annotated, Any

import pydantic
from pydantic import BaseModel, ConfigDict, Field, model_validator

from kpops.utils.docstring import describe_attr
from kpops.utils.pydantic import DescConfigModel


class OutputTopicTypes(str, Enum):
    """Types of output topic.

    - OUTPUT: output topic
    - ERROR: error topic
    """

    OUTPUT = "output"
    ERROR = "error"


class TopicConfig(DescConfigModel):
    """Configure an output topic.

    :param type: Topic type
    :param key_schema: Key schema class name
    :param value_schema: Value schema class name
    :param partitions_count: Number of partitions into which the topic is divided
    :param replication_factor: Replication factor of the topic
    :param configs: Topic configs
    :param role: Custom identifier belonging to one or multiple topics, provide only if `type` is `extra`
    """

    type: OutputTopicTypes | None = Field(
        default=None, title="Topic type", description=describe_attr("type", __doc__)
    )
    key_schema: str | None = Field(
        default=None,
        title="Key schema",
        description=describe_attr("key_schema", __doc__),
    )
    value_schema: str | None = Field(
        default=None,
        title="Value schema",
        description=describe_attr("value_schema", __doc__),
    )
    partitions_count: int | None = Field(
        default=None,
        title="Partitions count",
        description=describe_attr("partitions_count", __doc__),
    )
    replication_factor: int | None = Field(
        default=None,
        title="Replication factor",
        description=describe_attr("replication_factor", __doc__),
    )
    configs: dict[str, str | int] = Field(
        default={}, description=describe_attr("configs", __doc__)
    )
    role: str | None = Field(default=None, description=describe_attr("role", __doc__))

    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
        populate_by_name=True,
    )

    @model_validator(mode="after")
    def extra_topic_role(self) -> Any:
        """Ensure that `cls.role` is used correctly, assign type if needed."""
        if self.type and self.role:
            msg = "Define `role` only if `type` is undefined"
            raise ValueError(msg)
        return self


class KafkaTopic(BaseModel):
    """Internal representation of a Kafka topic.

    Users configure topic name and config through the `ToSection` of a `PipelineComponent`

    :param name: Topic name
    :param config: Topic config
    """

    name: str
    config: TopicConfig = TopicConfig()

    @property
    def id(self) -> str:
        """Unique identifier of this topic."""
        return f"topic-{self.name}"

    @staticmethod
    def deduplicate(topics: Iterable[KafkaTopic]) -> list[KafkaTopic]:
        """Deduplicate an iterable of Kafka topics based on their name, overwriting previous ones."""
        return list({topic.name: topic for topic in topics}.values())


def deserialize_kafka_topic_from_str(topic: Any) -> KafkaTopic | dict:
    if topic and isinstance(topic, str):
        return KafkaTopic(name=topic)
    if isinstance(topic, KafkaTopic | dict):
        return topic
    msg = f"Topic should be a valid {KafkaTopic.__name__} instance or topic name string"
    raise ValueError(msg)


def serialize_kafka_topic_to_str(topic: KafkaTopic) -> str:
    return topic.name


# Pydantic type for KafkaTopic that serializes to str, used for generating the correct JSON schema
KafkaTopicStr = Annotated[
    KafkaTopic,
    pydantic.WithJsonSchema({"type": "string"}),
    pydantic.BeforeValidator(deserialize_kafka_topic_from_str),
    pydantic.PlainSerializer(serialize_kafka_topic_to_str),
]



from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, TypeAlias

import networkx as nx
import yaml
from pydantic import (
    BaseModel,
    ConfigDict,
    SerializeAsAny,
    computed_field,
)

from kpops.components.base_components.pipeline_component import PipelineComponent
from kpops.utils.dict_ops import update_nested_pair
from kpops.utils.environment import ENV, PIPELINE_PATH
from kpops.utils.yaml import load_yaml_file

if TYPE_CHECKING:
    from collections.abc import Awaitable, Coroutine, Iterator
    from pathlib import Path

    from kpops.cli.registry import Registry
    from kpops.component_handlers import ComponentHandlers
    from kpops.config import KpopsConfig

log = logging.getLogger("PipelineGenerator")


class ParsingException(Exception):
    pass


class ValidationError(Exception):
    pass


ComponentFilterPredicate: TypeAlias = Callable[[PipelineComponent], bool]


class Pipeline(BaseModel):
    """Pipeline representation."""

    _component_index: dict[str, PipelineComponent] = {}
    _graph: nx.DiGraph = nx.DiGraph()

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @computed_field(title="Components")
    @property
    def components(self) -> list[SerializeAsAny[PipelineComponent]]:
        return list(self._component_index.values())

    @property
    def last(self) -> PipelineComponent:
        return self.components[-1]

    def add(self, component: PipelineComponent) -> None:
        if self._component_index.get(component.id) is not None:
            msg = (
                f"Pipeline steps must have unique id, '{component.id}' already exists."
            )
            raise ValidationError(msg)
        self._component_index[component.id] = component
        self.__add_to_graph(component)

    def remove(self, component_id: str) -> None:
        self._component_index.pop(component_id)

    def get(self, component_id: str) -> PipelineComponent | None:
        self._component_index.get(component_id)

    def find(self, predicate: ComponentFilterPredicate) -> Iterator[PipelineComponent]:
        """Find pipeline components matching a custom predicate.

        :param predicate: Filter function,
            returns boolean value whether the component should be kept or removed
        :returns: Iterator of components matching the predicate
        """
        for component in self.components:
            if predicate(component):
                yield component

    def filter(self, predicate: ComponentFilterPredicate) -> None:
        """Filter pipeline components using a custom predicate.

        :param predicate: Filter function,
            returns boolean value whether the component should be kept or removed
        """
        for component in self.components:
            # filter out components not matching the predicate
            if not predicate(component):
                self.remove(component.id)

    def validate(self) -> None:  # pyright: ignore [reportIncompatibleMethodOverride]
        self.__validate_graph()

    def to_yaml(self) -> str:
        return yaml.dump(
            self.model_dump(mode="json", by_alias=True, exclude_none=True)["components"]
        )

    def build_execution_graph(
        self,
        runner: Callable[[PipelineComponent], Coroutine[Any, Any, None]],
        /,
        reverse: bool = False,
    ) -> Awaitable[None]:
        async def run_parallel_tasks(
            coroutines: list[Coroutine[Any, Any, None]],
        ) -> None:
            tasks = []
            for coro in coroutines:
                tasks.append(asyncio.create_task(coro))
            await asyncio.gather(*tasks)

        async def run_graph_tasks(pending_tasks: list[Awaitable[None]]) -> None:
            for pending_task in pending_tasks:
                await pending_task

        graph: nx.DiGraph = self._graph.copy()  # pyright: ignore[reportAssignmentType, reportGeneralTypeIssues] imprecise type hint in networkx

        # We add an extra node to the graph, connecting all the leaf nodes to it
        # in that way we make this node the root of the graph, avoiding backtracking
        root_node = "root_node_bfs"
        graph.add_node(root_node)

        for node in graph:
            predecessors = list(graph.predecessors(node))
            if not predecessors:
                graph.add_edge(root_node, node)

        layers_graph: list[list[str]] = list(nx.bfs_layers(graph, root_node))

        sorted_tasks = []
        for layer in layers_graph[1:]:
            if parallel_tasks := self.__get_parallel_tasks_from(layer, runner):
                sorted_tasks.append(run_parallel_tasks(parallel_tasks))

        if reverse:
            sorted_tasks.reverse()

        return run_graph_tasks(sorted_tasks)

    def __getitem__(self, component_id: str) -> PipelineComponent:
        try:
            return self._component_index[component_id]
        except KeyError as exc:
            msg = f"Component {component_id} not found"
            raise ValueError(msg) from exc

    def __bool__(self) -> bool:
        return bool(self._component_index)

    def __iter__(self) -> Iterator[PipelineComponent]:  # pyright: ignore [reportIncompatibleMethodOverride]
        yield from self._component_index.values()

    def __len__(self) -> int:
        return len(self.components)

    def __add_to_graph(self, component: PipelineComponent):
        self._graph.add_node(component.id)

        for input_topic in component.inputs:
            self.__add_input(input_topic.id, component.id)

        for output_topic in component.outputs:
            self.__add_output(output_topic.id, component.id)

    def __add_output(self, topic_id: str, source: str) -> None:
        self._graph.add_node(topic_id)
        self._graph.add_edge(source, topic_id)

    def __add_input(self, topic_id: str, target: str) -> None:
        self._graph.add_node(topic_id)
        self._graph.add_edge(topic_id, target)

    def __get_parallel_tasks_from(
        self,
        layer: list[str],
        runner: Callable[[PipelineComponent], Coroutine[Any, Any, None]],
    ) -> list[Coroutine[Any, Any, None]]:
        def gen_parallel_tasks():
            for node_in_layer in layer:
                # check if component, skip topics
                if (component := self._component_index.get(node_in_layer)) is not None:
                    yield runner(component)

        return list(gen_parallel_tasks())

    def __validate_graph(self) -> None:
        if not nx.is_directed_acyclic_graph(self._graph):
            msg = "Pipeline is not a valid DAG."
            raise ValueError(msg)


def create_env_components_index(
    environment_components: list[dict[str, Any]],
) -> dict[str, dict[str, Any]]:
    """Create an index for all registered components in the project.

    :param environment_components: List of all components to be included
    :return: component index
    """
    index: dict[str, dict[str, Any]] = {}
    for component in environment_components:
        if "type" not in component or "name" not in component:
            msg = "To override components per environment, every component should at least have a type and a name."
            raise ValueError(msg)
        index[component["name"]] = component  # TODO: id
    return index


@dataclass
class PipelineGenerator:
    config: KpopsConfig
    registry: Registry
    handlers: ComponentHandlers
    pipeline: Pipeline = field(init=False, default_factory=Pipeline)

    def parse(
        self,
        components: list[dict],
        environment_components: list[dict],
    ) -> Pipeline:
        """Parse pipeline from sequence of component dictionaries.

        :param components: List of components
        :param environment_components: List of environment-specific components
        :returns: Initialized pipeline object
        """
        self.env_components_index = create_env_components_index(environment_components)
        self.parse_components(components)
        self.pipeline.validate()
        return self.pipeline

    def load_yaml(self, path: Path, environment: str | None) -> Pipeline:
        """Load pipeline definition from YAML file.

        The file is often named ``pipeline.yaml``

        :param path: Path to pipeline definition yaml file
        :param environment: Environment name
        :raises TypeError: The pipeline definition should contain a list of components
        :raises TypeError: The env-specific pipeline definition should contain a list of components
        :returns: Initialized pipeline object
        """
        PipelineGenerator.set_pipeline_name_env_vars(
            self.config.pipeline_base_dir, path
        )
        PipelineGenerator.set_environment_name(environment)
        PipelineGenerator.set_pipeline_path(path)

        main_content = load_yaml_file(path, substitution=ENV)
        if not isinstance(main_content, list):
            msg = f"The pipeline definition {path} should contain a list of components"
            raise TypeError(msg)
        env_content = []
        if (
            environment
            and (
                env_file := PipelineGenerator.pipeline_filename_environment(
                    path, environment
                )
            ).exists()
        ):
            env_content = load_yaml_file(env_file, substitution=ENV)
            if not isinstance(env_content, list):
                msg = f"The pipeline definition {env_file} should contain a list of components"
                raise TypeError(msg)

        return self.parse(main_content, env_content)

    def parse_components(self, components: list[dict]) -> None:
        """Instantiate, enrich and inflate a list of components.

        :param components: List of components
        :raises ValueError: Every component must have a type defined
        :raises ParsingException: Error enriching component
        :raises ParsingException: All undefined exceptions
        """
        for component_data in components:
            try:
                try:
                    component_type: str = component_data["type"]
                except KeyError as ke:
                    msg = "Every component must have a type defined, this component does not have one."
                    raise ValueError(msg) from ke
                component_class = self.registry[component_type]
                self.apply_component(component_class, component_data)
            except Exception as ex:  # noqa: BLE001
                if "name" in component_data:
                    msg = f"Error enriching {component_data['type']} component {component_data['name']}"
                    raise ParsingException(msg) from ex
                else:
                    raise ParsingException from ex

    def apply_component(
        self, component_class: type[PipelineComponent], component_data: dict[str, Any]
    ) -> None:
        """Instantiate, enrich and inflate pipeline component.

        Applies input topics according to FromSection.

        :param component_class: Type of pipeline component
        :param component_data: Arguments for instantiation of pipeline component
        """
        component = component_class(
            config=self.config,
            handlers=self.handlers,
            **component_data,
        )
        component = self.enrich_component_with_env(component)
        # inflate & enrich components
        for inflated_component in component.inflate():  # TODO: recursively
            if inflated_component.from_:
                # read from specified components
                for (
                    original_from_component_id,
                    from_topic,
                ) in inflated_component.from_.components.items():
                    original_from_component = self.pipeline[original_from_component_id]

                    inflated_from_component = original_from_component.inflate()[-1]
                    resolved_from_component = self.pipeline[inflated_from_component.id]

                    inflated_component.weave_from_topics(
                        resolved_from_component.to, from_topic
                    )
            elif self.pipeline:
                # read from previous component
                prev_component = self.pipeline.last
                inflated_component.weave_from_topics(prev_component.to)
            self.pipeline.add(inflated_component)

    def enrich_component_with_env(
        self, component: PipelineComponent
    ) -> PipelineComponent:
        """Enrich a pipeline component with env-specific config.

        :param component: Component to be enriched
        :returns: Enriched component
        """
        env_component = self.env_components_index.get(component.name)
        if not env_component:
            return component
        env_component_as_dict = update_nested_pair(
            env_component,
            component.model_dump(mode="json", by_alias=True),
        )

        return component.__class__(
            config=self.config,
            handlers=self.handlers,
            **env_component_as_dict,
        )

    @staticmethod
    def pipeline_filename_environment(pipeline_path: Path, environment: str) -> Path:
        """Add the environment name from the KpopsConfig to the pipeline.yaml path.

        :param pipeline_path: Path to pipeline.yaml file
        :param environment: Environment name
        :returns: An absolute path to the pipeline_<environment>.yaml
        """
        return pipeline_path.with_stem(f"{pipeline_path.stem}_{environment}")

    @staticmethod
    def set_pipeline_name_env_vars(
        pipeline_base_dir: Path, pipeline_path: Path
    ) -> None:
        """Set the environment variable pipeline_name relative to the given base_dir.

        Moreover, for each sub-path an environment variable is set.
        For example, for a given path ./data/v1/dev/pipeline.yaml the pipeline_name would be
        set to data-v1-dev. Then the sub environment variables are set:

        .. code-block:: python
            pipeline.name_0 = data
            pipeline.name_1 = v1
            pipeline.name_2 = dev

        :param pipeline_base_dir: Base directory to the pipeline files
        :param pipeline_path: Path to pipeline.yaml file
        """
        path_without_file = (
            pipeline_path.resolve().relative_to(pipeline_base_dir.resolve()).parts[:-1]
        )
        if not path_without_file:
            msg = "The pipeline-base-dir should not equal the pipeline-path"
            raise ValueError(msg)
        pipeline_name = "-".join(path_without_file)
        ENV["pipeline.name"] = pipeline_name
        for level, parent in enumerate(path_without_file):
            ENV[f"pipeline.name_{level}"] = parent

    @staticmethod
    def set_environment_name(environment: str | None) -> None:
        """Set the environment name.

        It will be used to find environment-specific pipeline definitions,
        defaults and configs.

        :param environment: Environment name
        """
        if environment is not None:
            ENV["environment"] = environment

    @staticmethod
    def set_pipeline_path(path: Path):
        ENV[PIPELINE_PATH] = str(path.resolve())



from __future__ import annotations

import logging
from abc import ABC
from functools import cached_property
from typing import Any, NoReturn

from pydantic import Field, PrivateAttr, ValidationInfo, computed_field, field_validator
from typing_extensions import override

from kpops.component_handlers.helm_wrapper.model import (
    HelmRepoConfig,
)
from kpops.component_handlers.kafka_connect.model import (
    KafkaConnectorConfig,
    KafkaConnectorResetterConfig,
    KafkaConnectorResetterValues,
    KafkaConnectorType,
)
from kpops.components.base_components.cleaner import Cleaner
from kpops.components.base_components.helm_app import HelmAppValues
from kpops.components.base_components.models.from_section import FromTopic
from kpops.components.base_components.models.topic import KafkaTopic
from kpops.components.base_components.pipeline_component import PipelineComponent
from kpops.utils.colorify import magentaify
from kpops.utils.docstring import describe_attr

log = logging.getLogger("KafkaConnector")


class KafkaConnectorResetter(Cleaner, ABC):
    """Helm app for resetting and cleaning a Kafka Connector.

    :param repo_config: Configuration of the Helm chart repo to be used for
        deploying the component, defaults to kafka-connect-resetter Helm repo
    :param version: Helm chart version, defaults to "1.0.4"
    """

    from_: None = None
    to: None = None
    app: KafkaConnectorResetterValues
    repo_config: HelmRepoConfig = Field(
        default=HelmRepoConfig(
            repository_name="bakdata-kafka-connect-resetter",
            url="https://bakdata.github.io/kafka-connect-resetter/",
        )
    )
    version: str | None = Field(
        default="1.0.4", description=describe_attr("version", __doc__)
    )

    @property
    @override
    def helm_chart(self) -> str:
        return f"{self.repo_config.repository_name}/kafka-connect-resetter"

    @override
    async def reset(self, dry_run: bool) -> None:
        """Reset connector.

        At first, it deletes the previous cleanup job (connector resetter)
        to make sure that there is no running clean job in the cluster. Then it releases a cleanup job.
        If retain_clean_jobs config is set to false the cleanup job will be deleted subsequently.

        :param dry_run: If the cleanup should be run in dry run mode or not
        """
        log.info(
            magentaify(
                f"Connector Cleanup: uninstalling cleanup job Helm release from previous runs for {self.app.config.connector}"
            )
        )
        await self.destroy(dry_run)

        log.info(
            magentaify(
                f"Connector Cleanup: deploy Connect {self.app.connector_type} resetter for {self.app.config.connector}"
            )
        )
        await self.deploy(dry_run)

        if not self.config.retain_clean_jobs:
            log.info(magentaify("Connector Cleanup: uninstall Kafka Resetter."))
            await self.destroy(dry_run)

    @override
    async def clean(self, dry_run: bool) -> None:
        await self.reset(dry_run)


class KafkaConnector(PipelineComponent, ABC):
    """Base class for all Kafka connectors.

    Should only be used to set defaults

    :param app: Application-specific settings
    :param resetter_namespace: Kubernetes namespace in which the Kafka Connect resetter shall be deployed
    :param resetter_values: Overriding Kafka Connect resetter Helm values, e.g. to override the image tag etc.,
        defaults to empty HelmAppValues
    """

    app: KafkaConnectorConfig = Field(
        default=...,
        description=describe_attr("app", __doc__),
    )
    resetter_namespace: str | None = Field(
        default=None, description=describe_attr("resetter_namespace", __doc__)
    )
    resetter_values: HelmAppValues = Field(
        default_factory=HelmAppValues,
        description=describe_attr("resetter_values", __doc__),
    )
    _connector_type: KafkaConnectorType = PrivateAttr()

    @field_validator("app", mode="before")
    @classmethod
    def connector_config_should_have_component_name(
        cls,
        app: KafkaConnectorConfig | dict[str, Any],
        info: ValidationInfo,
    ) -> KafkaConnectorConfig:
        if isinstance(app, KafkaConnectorConfig):
            app = app.model_dump()
        component_name: str = info.data["prefix"] + info.data["name"]
        connector_name: str | None = app.get("name")
        if connector_name is not None and connector_name != component_name:
            msg = f"Connector name '{connector_name}' should be the same as component name '{component_name}'"
            raise ValueError(msg)
        app["name"] = component_name
        return KafkaConnectorConfig(**app)

    @cached_property
    def _resetter(self) -> KafkaConnectorResetter:
        kwargs: dict[str, Any] = {}
        if self.resetter_namespace:
            kwargs["namespace"] = self.resetter_namespace
        return KafkaConnectorResetter(
            config=self.config,
            handlers=self.handlers,
            **kwargs,
            **self.model_dump(
                by_alias=True,
                exclude={
                    "_resetter",
                    "resetter_values",
                    "resetter_namespace",
                    "app",
                    "from_",
                    "to",
                },
            ),
            app=KafkaConnectorResetterValues(
                connector_type=self._connector_type.value,
                config=KafkaConnectorResetterConfig(
                    connector=self.full_name,
                    brokers=self.config.kafka_brokers,
                ),
                **self.resetter_values.model_dump(),
            ),
        )

    @override
    async def deploy(self, dry_run: bool) -> None:
        if self.to:
            for topic in self.to.kafka_topics:
                await self.handlers.topic_handler.create_topic(topic, dry_run=dry_run)

            if self.handlers.schema_handler:
                await self.handlers.schema_handler.submit_schemas(
                    to_section=self.to, dry_run=dry_run
                )

        await self.handlers.connector_handler.create_connector(
            self.app, dry_run=dry_run
        )

    @override
    async def destroy(self, dry_run: bool) -> None:
        await self.handlers.connector_handler.destroy_connector(
            self.full_name, dry_run=dry_run
        )

    @override
    async def clean(self, dry_run: bool) -> None:
        if self.to:
            if self.handlers.schema_handler:
                await self.handlers.schema_handler.delete_schemas(
                    to_section=self.to, dry_run=dry_run
                )
            for topic in self.to.kafka_topics:
                await self.handlers.topic_handler.delete_topic(topic, dry_run=dry_run)


class KafkaSourceConnector(KafkaConnector):
    """Kafka source connector model.

    :param offset_topic: `offset.storage.topic`,
        more info: https://kafka.apache.org/documentation/#connect_running,
        defaults to None
    """

    offset_topic: str | None = Field(
        default=None,
        description=describe_attr("offset_topic", __doc__),
    )

    _connector_type: KafkaConnectorType = PrivateAttr(KafkaConnectorType.SOURCE)

    @computed_field
    @cached_property
    def _resetter(self) -> KafkaConnectorResetter:
        return super()._resetter

    @override
    def apply_from_inputs(self, name: str, topic: FromTopic) -> NoReturn:
        msg = "Kafka source connector doesn't support FromSection"
        raise NotImplementedError(msg)

    @override
    async def reset(self, dry_run: bool) -> None:
        self._resetter.app.config.offset_topic = self.offset_topic
        await self._resetter.reset(dry_run)

    @override
    async def clean(self, dry_run: bool) -> None:
        await super().clean(dry_run)
        self._resetter.app.config.offset_topic = self.offset_topic
        await self._resetter.clean(dry_run)


class KafkaSinkConnector(KafkaConnector):
    """Kafka sink connector model."""

    _connector_type: KafkaConnectorType = PrivateAttr(KafkaConnectorType.SINK)

    @computed_field
    @cached_property
    def _resetter(self) -> KafkaConnectorResetter:
        return super()._resetter

    @property
    @override
    def input_topics(self) -> list[KafkaTopic]:
        return self.app.topics

    @override
    def add_input_topics(self, topics: list[KafkaTopic]) -> None:
        self.app.topics = KafkaTopic.deduplicate(self.app.topics + topics)

    @override
    def set_input_pattern(self, name: str) -> None:
        self.app.topics_regex = name

    @override
    def set_error_topic(self, topic: KafkaTopic) -> None:
        self.app.errors_deadletterqueue_topic_name = topic

    @override
    async def reset(self, dry_run: bool) -> None:
        self._resetter.app.config.delete_consumer_group = False
        await self._resetter.reset(dry_run)

    @override
    async def clean(self, dry_run: bool) -> None:
        await super().clean(dry_run)
        self._resetter.app.config.delete_consumer_group = True
        await self._resetter.clean(dry_run)



from kpops.component_handlers.kubernetes.model import K8S_LABEL_MAX_LEN
from kpops.component_handlers.kubernetes.utils import trim

RELEASE_NAME_MAX_LEN = 53


def create_helm_release_name(name: str, suffix: str = "") -> str:
    """Shortens the long Helm release name.

    :param name: The Helm release name to be shortened.
    :param suffix: The release suffix to preserve
    :return: Trimmed + hashed version of the release name if it exceeds the Helm release character length, otherwise the input name
    """
    return trim(RELEASE_NAME_MAX_LEN, name, suffix)


def create_helm_name_override(name: str, suffix: str = "") -> str:
    """Create Helm chart name override.

    :param name: The Helm name override to be shortened.
    :param suffix: The release suffix to preserve
    :return: Trimmed + hashed version of the name override if it exceeds the Kubernetes character length, otherwise the input name
    """
    return trim(K8S_LABEL_MAX_LEN, name, suffix)



from __future__ import annotations

import logging
from functools import cached_property
from typing import Annotated, Any

import pydantic
from pydantic import Field, SerializationInfo, computed_field, model_serializer
from typing_extensions import override

from kpops.component_handlers.helm_wrapper.dry_run_handler import DryRunHandler
from kpops.component_handlers.helm_wrapper.helm import Helm
from kpops.component_handlers.helm_wrapper.helm_diff import HelmDiff
from kpops.component_handlers.helm_wrapper.model import (
    HelmFlags,
    HelmRepoConfig,
    HelmTemplateFlags,
    HelmUpgradeInstallFlags,
)
from kpops.component_handlers.helm_wrapper.utils import (
    create_helm_name_override,
    create_helm_release_name,
)
from kpops.component_handlers.kubernetes.model import K8S_LABEL_MAX_LEN
from kpops.components.base_components.kubernetes_app import (
    KubernetesApp,
    KubernetesAppValues,
)
from kpops.components.base_components.models.resource import Resource
from kpops.utils.colorify import magentaify
from kpops.utils.docstring import describe_attr
from kpops.utils.pydantic import exclude_by_name

log = logging.getLogger("HelmApp")


class HelmAppValues(KubernetesAppValues):
    """Helm app values.

    :param name_override: Helm chart name override, assigned automatically
    """

    name_override: (
        Annotated[str, pydantic.StringConstraints(max_length=K8S_LABEL_MAX_LEN)] | None
    ) = Field(
        default=None,
        title="Nameoverride",
        description=describe_attr("name_override", __doc__),
    )

    # TODO(Ivan Yordanov): Replace with a function decorated with `@model_serializer`
    # BEWARE! All default values are enforced, hard to replicate without
    # access to ``model_dump``
    @override
    def model_dump(self, **_) -> dict[str, Any]:
        return super().model_dump(
            by_alias=True, exclude_none=True, exclude_defaults=True
        )


class HelmApp(KubernetesApp):
    """Kubernetes app managed through Helm with an associated Helm chart.

    :param repo_config: Configuration of the Helm chart repo to be used for
        deploying the component, defaults to None this means that the command "helm repo add" is not called and Helm
        expects a path to local Helm chart.
    :param version: Helm chart version, defaults to None
    :param app: Helm app values
    """

    repo_config: HelmRepoConfig | None = Field(
        default=None,
        description=describe_attr("repo_config", __doc__),
    )
    version: str | None = Field(
        default=None,
        description=describe_attr("version", __doc__),
    )
    app: HelmAppValues = Field(
        default=...,
        description=describe_attr("app", __doc__),
    )

    @cached_property
    def helm(self) -> Helm:
        """Helm object that contains component-specific config such as repo."""
        helm = Helm(self.config.helm_config)
        if self.repo_config is not None:
            helm.add_repo(
                self.repo_config.repository_name,
                self.repo_config.url,
                self.repo_config.repo_auth_flags,
            )
        return helm

    @cached_property
    def helm_diff(self) -> HelmDiff:
        """Helm diff object of last and current release of this component."""
        return HelmDiff(self.config.helm_diff_config)

    @cached_property
    def dry_run_handler(self) -> DryRunHandler:
        helm_diff = HelmDiff(self.config.helm_diff_config)
        return DryRunHandler(self.helm, helm_diff, self.namespace)

    @computed_field
    @property
    def helm_release_name(self) -> str:
        """The name for the Helm release."""
        return create_helm_release_name(self.full_name)

    @computed_field
    @property
    def helm_name_override(self) -> str:
        """Helm chart name override."""
        return create_helm_name_override(self.full_name)

    @property
    def helm_chart(self) -> str:
        """Return component's Helm chart."""
        msg = (
            f"Please implement the helm_chart property of the {self.__module__} module."
        )
        raise NotImplementedError(msg)

    @property
    def helm_flags(self) -> HelmFlags:
        """Return shared flags for Helm commands."""
        auth_flags = (
            self.repo_config.repo_auth_flags.model_dump() if self.repo_config else {}
        )
        return HelmFlags(
            **auth_flags,
            version=self.version,
            create_namespace=self.config.create_namespace,
        )

    @property
    def template_flags(self) -> HelmTemplateFlags:
        """Return flags for Helm template command."""
        return HelmTemplateFlags(
            **self.helm_flags.model_dump(),
            api_version=self.config.helm_config.api_version,
        )

    @override
    def manifest(self) -> Resource:
        return self.helm.template(
            self.helm_release_name,
            self.helm_chart,
            self.namespace,
            self.to_helm_values(),
            self.template_flags,
        )

    @property
    def deploy_flags(self) -> HelmUpgradeInstallFlags:
        """Return flags for Helm upgrade install command."""
        return HelmUpgradeInstallFlags(**self.helm_flags.model_dump())

    @override
    async def deploy(self, dry_run: bool) -> None:
        stdout = await self.helm.upgrade_install(
            self.helm_release_name,
            self.helm_chart,
            dry_run,
            self.namespace,
            self.to_helm_values(),
            self.deploy_flags,
        )
        if dry_run:
            self.dry_run_handler.print_helm_diff(stdout, self.helm_release_name, log)

    @override
    async def destroy(self, dry_run: bool) -> None:
        stdout = await self.helm.uninstall(
            self.namespace,
            self.helm_release_name,
            dry_run,
        )

        if stdout:
            log.info(magentaify(stdout))

    def to_helm_values(self) -> dict[str, Any]:
        """Generate a dictionary of values readable by Helm from `self.app`.

        :returns: The values to be used by Helm
        """
        if self.app.name_override is None:
            self.app.name_override = self.helm_name_override
        return self.app.model_dump()

    def print_helm_diff(self, stdout: str) -> None:
        """Print the diff of the last and current release of this component.

        :param stdout: The output of a Helm command that installs or upgrades the release
        """
        current_release = list(
            self.helm.get_manifest(self.helm_release_name, self.namespace)
        )
        if current_release:
            log.info(f"Helm release {self.helm_release_name} already exists")
        else:
            log.info(f"Helm release {self.helm_release_name} does not exist")
        new_release = Helm.load_manifest(stdout)
        self.helm_diff.log_helm_diff(log, current_release, new_release)

    # HACK: workaround for Pydantic to exclude cached properties during model export
    # TODO(Ivan Yordanov): Currently hacky and potentially unsafe. Find cleaner solution
    @model_serializer(mode="wrap", when_used="always")
    def serialize_model(self, handler, info: SerializationInfo) -> dict[str, Any]:
        return exclude_by_name(handler(self), "helm", "helm_diff")



from __future__ import annotations

import logging
from functools import cached_property
from typing import TYPE_CHECKING

import httpx

from kpops.component_handlers.topic.exception import (
    KafkaRestProxyError,
    TopicNotFoundException,
)
from kpops.component_handlers.topic.model import (
    BrokerConfigResponse,
    TopicConfigResponse,
    TopicResponse,
    TopicSpec,
)

if TYPE_CHECKING:
    from pydantic import AnyHttpUrl

    from kpops.config import KafkaRestConfig

log = logging.getLogger("KafkaRestProxy")

HEADERS = {"Content-Type": "application/json"}


class ProxyWrapper:
    """Wraps Kafka REST Proxy APIs."""

    def __init__(self, config: KafkaRestConfig) -> None:
        self._config: KafkaRestConfig = config
        self._client = httpx.AsyncClient(timeout=config.timeout)
        self._sync_client = httpx.Client(timeout=config.timeout)

    @cached_property
    def cluster_id(self) -> str:
        """Get the Kafka cluster ID by sending a request to Kafka REST proxy.

        More information about the cluster ID can be found here:
        https://docs.confluent.io/platform/current/kafka-rest/api.html#cluster-v3.

        Currently both Kafka and Kafka REST Proxy are only aware of the Kafka cluster pointed at by the
        bootstrap.servers configuration. Therefore, only one Kafka cluster will be returned.

        :raises KafkaRestProxyError: Kafka REST proxy error
        :return: The Kafka cluster ID.
        """
        response = self._sync_client.get(url=f"{self._config.url!s}v3/clusters")

        if response.status_code == httpx.codes.OK:
            cluster_information = response.json()
            return cluster_information["data"][0]["cluster_id"]

        raise KafkaRestProxyError(response)

    @property
    def url(self) -> AnyHttpUrl:
        return self._config.url

    async def create_topic(self, topic_spec: TopicSpec) -> None:
        """Create a topic.

        API Reference:
        https://docs.confluent.io/platform/current/kafka-rest/api.html#post--clusters-cluster_id-topics

        :param topic_spec: The topic specification.
        :raises KafkaRestProxyError: Kafka REST proxy error
        """
        response = await self._client.post(
            url=f"{self.url!s}v3/clusters/{self.cluster_id}/topics",
            headers=HEADERS,
            json=topic_spec.model_dump(exclude_none=True),
        )

        if response.status_code == httpx.codes.CREATED:
            log.info(f"Topic {topic_spec.topic_name} created.")
            log.debug(response.json())
            return

        raise KafkaRestProxyError(response)

    async def delete_topic(self, topic_name: str) -> None:
        """Delete a topic.

        API Reference:
        https://docs.confluent.io/platform/current/kafka-rest/api.html#delete--clusters-cluster_id-topics-topic_name

        :param topic_name: Name of the topic.
        :raises KafkaRestProxyError: Kafka REST proxy error
        """
        response = await self._client.delete(
            url=f"{self.url!s}v3/clusters/{self.cluster_id}/topics/{topic_name}",
            headers=HEADERS,
        )

        if response.status_code == httpx.codes.NO_CONTENT:
            log.info(f"Topic {topic_name} deleted.")
            return

        raise KafkaRestProxyError(response)

    async def get_topic(self, topic_name: str) -> TopicResponse:
        """Return the topic with the given topic_name.

        API Reference:
        https://docs.confluent.io/platform/current/kafka-rest/api.html#get--clusters-cluster_id-topics-topic_name
        :param topic_name: The topic name.
        :raises TopicNotFoundException: Topic not found
        :raises KafkaRestProxyError: Kafka REST proxy error
        :return: Response of the get topic API.
        """
        response = await self._client.get(
            url=f"{self.url!s}v3/clusters/{self.cluster_id}/topics/{topic_name}",
            headers=HEADERS,
        )

        if response.status_code == httpx.codes.OK:
            log.debug(f"Topic {topic_name} found.")
            log.debug(response.json())
            return TopicResponse(**response.json())

        elif (
            response.status_code == httpx.codes.NOT_FOUND
            and response.json()["error_code"] == 40403
        ):
            log.debug(f"Topic {topic_name} not found.")
            log.debug(response.json())
            raise TopicNotFoundException

        raise KafkaRestProxyError(response)

    async def get_topic_config(self, topic_name: str) -> TopicConfigResponse:
        """Return the config with the given topic_name.

        API Reference:
        https://docs.confluent.io/platform/current/kafka-rest/api.html#acl-v3
        :param topic_name: The topic name.
        :raises TopicNotFoundException: Topic not found
        :raises KafkaRestProxyError: Kafka REST proxy error
        :return: The topic configuration.
        """
        response = await self._client.get(
            url=f"{self.url!s}v3/clusters/{self.cluster_id}/topics/{topic_name}/configs",
            headers=HEADERS,
        )

        if response.status_code == httpx.codes.OK:
            log.debug(f"Configs for {topic_name} found.")
            log.debug(response.json())
            return TopicConfigResponse(**response.json())

        elif (
            response.status_code == httpx.codes.NOT_FOUND
            and response.json()["error_code"] == 40403
        ):
            log.debug(f"Configs for {topic_name} not found.")
            log.debug(response.json())
            raise TopicNotFoundException

        raise KafkaRestProxyError(response)

    async def batch_alter_topic_config(
        self, topic_name: str, json_body: list[dict]
    ) -> None:
        """Reset config of given config_name param to the default value on the Kafka server.

        API Reference:
        https://docs.confluent.io/platform/current/kafka-rest/api.html#post--clusters-cluster_id-topics-topic_name-configs-alter

        :param topic_name: The topic name.
        :param config_name: The configuration parameter name.
        :raises KafkaRestProxyError: Kafka REST proxy error
        """
        response = await self._client.post(
            url=f"{self.url!s}v3/clusters/{self.cluster_id}/topics/{topic_name}/configs:alter",
            headers=HEADERS,
            json={"data": json_body},
        )

        if response.status_code == httpx.codes.NO_CONTENT:
            log.info(f"Config of topic {topic_name} was altered.")
            return

        raise KafkaRestProxyError(response)

    async def get_broker_config(self) -> BrokerConfigResponse:
        """Return the list of configuration parameters for all the brokers in the given Kafka cluster.

        API Reference:
        https://docs.confluent.io/platform/current/kafka-rest/api.html#get--clusters-cluster_id-brokers---configs

        :raises KafkaRestProxyError: Kafka REST proxy error
        :return: The broker configuration.
        """
        response = await self._client.get(
            url=f"{self.url!s}v3/clusters/{self.cluster_id}/brokers/-/configs",
            headers=HEADERS,
        )

        if response.status_code == httpx.codes.OK:
            log.debug("Broker configs found.")
            log.debug(response.json())
            return BrokerConfigResponse(**response.json())

        raise KafkaRestProxyError(response)



from __future__ import annotations

import json
import logging
from abc import ABC
from collections.abc import Hashable, Sequence
from dataclasses import asdict, is_dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, TypeVar

import pydantic
import typer
from pydantic import (
    AliasChoices,
    ConfigDict,
    Field,
    computed_field,
)
from pydantic.json_schema import SkipJsonSchema

from kpops.component_handlers import ComponentHandlers
from kpops.config import KpopsConfig
from kpops.utils import cached_classproperty
from kpops.utils.dict_ops import (
    generate_substitution,
    update_nested,
    update_nested_pair,
)
from kpops.utils.docstring import describe_attr
from kpops.utils.environment import ENV, PIPELINE_PATH
from kpops.utils.pydantic import DescConfigModel, issubclass_patched, to_dash
from kpops.utils.types import JsonType
from kpops.utils.yaml import load_yaml_file, substitute_nested

try:
    from typing import Self  # pyright: ignore[reportAttributeAccessIssue]
except ImportError:
    from typing_extensions import Self

log = logging.getLogger("BaseDefaultsComponent")


class BaseDefaultsComponent(DescConfigModel, ABC):
    """Base for all components, handles defaults.

    Component defaults are usually provided in a YAML file called
    `defaults.yaml`. This class ensures that the defaults are read and assigned
    correctly to the component.

    :param enrich: Whether to enrich component with defaults, defaults to False
    :param config: KPOps configuration to be accessed by this component
    :param handlers: Component handlers to be accessed by this component
    :param validate: Whether to run custom validation on the component, defaults to True
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        ignored_types=(cached_property, cached_classproperty),
    )
    enrich: SkipJsonSchema[bool] = Field(
        default=True,
        description=describe_attr("enrich", __doc__),
        exclude=True,
    )
    config: SkipJsonSchema[KpopsConfig] = Field(
        default=...,
        description=describe_attr("config", __doc__),
        exclude=True,
    )
    handlers: SkipJsonSchema[ComponentHandlers] = Field(
        default=...,
        description=describe_attr("handlers", __doc__),
        exclude=True,
    )
    validate_: SkipJsonSchema[bool] = Field(
        validation_alias=AliasChoices("validate", "validate_"),
        default=False,
        description=describe_attr("validate", __doc__),
        exclude=True,
    )

    def __init__(self, **values: Any) -> None:
        if values.get("enrich", True):
            cls = self.__class__
            values = cls.extend_with_defaults(**values)
            tmp_self = cls(**values, enrich=False)
            values = tmp_self.model_dump(mode="json", by_alias=True)
            values = cls.substitute_in_component(tmp_self.config, **values)
            self.__init__(
                enrich=False,
                validate=True,
                config=tmp_self.config,
                handlers=tmp_self.handlers,
                **values,
            )
        else:
            super().__init__(**values)

    @pydantic.model_validator(mode="after")
    def validate_component(self) -> Self:
        if not self.enrich and self.validate_:
            self._validate_custom()
        return self

    @computed_field
    @cached_classproperty
    def type(cls: type[Self]) -> str:  # pyright: ignore[reportGeneralTypeIssues]
        """Return calling component's type.

        :returns: Component class name in dash-case
        """
        return to_dash(cls.__name__)

    @cached_classproperty
    def parents(cls: type[Self]) -> tuple[type[BaseDefaultsComponent], ...]:  # pyright: ignore[reportGeneralTypeIssues]
        """Get parent components.

        :return: All ancestor KPOps components
        """

        def gen_parents():
            for base in cls.mro():
                # skip class itself and non-component ancestors
                if base is cls or not issubclass_patched(base, BaseDefaultsComponent):
                    continue
                yield base

        return tuple(gen_parents())

    @classmethod
    def substitute_in_component(
        cls, config: KpopsConfig, **component_data: Any
    ) -> dict[str, Any]:
        """Substitute all $-placeholders in a component in dict representation.

        :param component_as_dict: Component represented as dict
        :return: Updated component
        """
        # Leftover variables that were previously introduced in the component by the substitution
        # functions, still hardcoded, because of their names.
        # TODO(Ivan Yordanov): Get rid of them
        substitution_hardcoded: dict[str, JsonType] = {
            "error_topic_name": config.topic_name_config.default_error_topic_name,
            "output_topic_name": config.topic_name_config.default_output_topic_name,
        }
        component_substitution = generate_substitution(
            component_data,
            "component",
            substitution_hardcoded,
            separator=".",
        )
        substitution = generate_substitution(
            config.model_dump(mode="json"),
            "config",
            existing_substitution=component_substitution,
            separator=".",
        )

        return json.loads(
            substitute_nested(
                json.dumps(component_data),
                **update_nested_pair(substitution, ENV),
            )
        )

    @classmethod
    def extend_with_defaults(cls, config: KpopsConfig, **kwargs: Any) -> dict[str, Any]:
        """Merge parent components' defaults with own.

        :param config: KPOps configuration
        :param kwargs: The init kwargs for pydantic
        :returns: Enriched kwargs with inherited defaults
        """
        pipeline_path_str = ENV.get(PIPELINE_PATH)
        kwargs["config"] = config
        if not pipeline_path_str:
            return kwargs
        pipeline_path = Path(pipeline_path_str)
        for k, v in kwargs.items():
            if isinstance(v, pydantic.BaseModel):
                kwargs[k] = v.model_dump(exclude_unset=True)
            elif is_dataclass(v):
                kwargs[k] = asdict(v)

        defaults_file_paths_ = get_defaults_file_paths(
            pipeline_path, config, ENV.get("environment")
        )
        defaults = cls.load_defaults(*defaults_file_paths_)
        log.debug(
            typer.style("Enriching component of type ", bold=False)
            + typer.style(cls.type, bold=True, underline=True)
        )
        return update_nested_pair(kwargs, defaults)

    @classmethod
    def load_defaults(cls, *defaults_file_paths: Path) -> dict[str, Any]:
        """Resolve component-specific defaults including environment defaults.

        :param defaults_file_paths: Path to `defaults.yaml`, ordered from highest to lowest priority,
         i.e. `defaults.yaml`, `defaults_{environment}`.yaml
        :returns: Component defaults
        """
        defaults: dict[str, Any] = {}
        for base in (cls, *cls.parents):
            component_type: str = base.type
            defaults = update_nested(
                defaults,
                *(
                    defaults_from_yaml(path, component_type)
                    for path in defaults_file_paths
                    if path.exists()
                ),
            )
        return defaults

    def _validate_custom(self) -> None:
        """Run custom validation on component."""


def defaults_from_yaml(path: Path, key: str) -> dict:
    """Read component-specific settings from a ``defaults*.yaml`` file and return @default if not found.

    :param path: Path to ``defaults*.yaml`` file
    :param key: Component type
    :returns: All defaults set for the given component in the provided YAML

    :Example:
    .. code-block:: python
        kafka_app_defaults = defaults_from_yaml(Path("/path/to/defaults.yaml"), "kafka-app")
    """
    content = load_yaml_file(path, substitution=ENV)
    if not isinstance(content, dict):
        msg = (
            "Default files should be structured as map ([app type] -> [default config]"
        )
        raise TypeError(msg)
    value = content.get(key)
    if value is None:
        return {}
    default_path = path.relative_to(Path.cwd())
    log.debug(
        f"Found defaults for component type {typer.style(key, bold=True, fg=typer.colors.MAGENTA)} in {default_path}"
    )
    return value


def get_defaults_file_paths(
    pipeline_path: Path, config: KpopsConfig, environment: str | None
) -> list[Path]:
    """Return a list of default file paths related to the given pipeline.

    This function traverses the directory hierarchy upwards till the `pipeline_base_dir`,
    starting from the directory containing
    the pipeline.yaml file specified by `pipeline_path`, looking for default files
    associated with the pipeline.

    :param pipeline_path: The path to the pipeline.yaml file.
    :param config: The KPOps configuration object containing settings such as pipeline_base_dir
                   and defaults_filename_prefix.
    :param environment: Optional. The environment for which default configuration files are sought.
    :returns: A list of Path objects representing the default configuration file paths.
    """
    default_paths = []

    if not pipeline_path.is_file():
        message = f"{pipeline_path} is not a valid pipeline file."
        raise FileNotFoundError(message)

    path = pipeline_path.resolve()
    pipeline_base_dir = config.pipeline_base_dir.resolve()
    if pipeline_base_dir not in path.parents:
        message = f"The given pipeline base path {pipeline_base_dir} is not part of the pipeline path {path}"
        raise RuntimeError(message)
    while pipeline_base_dir != path:
        environment_default_file_path = (
            path.parent / f"{config.defaults_filename_prefix}_{environment}.yaml"
        )
        if environment_default_file_path.is_file():
            default_paths.append(environment_default_file_path)

        defaults_yaml_path = path.parent / f"{config.defaults_filename_prefix}.yaml"
        if defaults_yaml_path.is_file():
            default_paths.append(defaults_yaml_path)

        path = path.parent

    return default_paths


_T = TypeVar("_T", bound=Hashable)


def deduplicate(seq: Sequence[_T]) -> list[_T]:
    """Deduplicate items of a sequence while preserving its order.

    :param seq: Sequence to be 'cleaned'
    :returns: Cleaned sequence in the form of a list
    """
    return list(dict.fromkeys(seq))



from functools import cached_property

from pydantic import Field, computed_field
from typing_extensions import override

from kpops.components.base_components.kafka_app import (
    KafkaApp,
    KafkaAppCleaner,
)
from kpops.components.base_components.models.topic import (
    KafkaTopic,
    OutputTopicTypes,
    TopicConfig,
)
from kpops.components.streams_bootstrap import StreamsBootstrap
from kpops.components.streams_bootstrap.app_type import AppType
from kpops.components.streams_bootstrap.producer.model import ProducerAppValues
from kpops.utils.docstring import describe_attr


class ProducerAppCleaner(KafkaAppCleaner):
    app: ProducerAppValues

    @property
    @override
    def helm_chart(self) -> str:
        return (
            f"{self.repo_config.repository_name}/{AppType.CLEANUP_PRODUCER_APP.value}"
        )


class ProducerApp(KafkaApp, StreamsBootstrap):
    """Producer component.

    This producer holds configuration to use as values for the streams-bootstrap
    producer Helm chart.

    Note that the producer does not support error topics.

    :param app: Application-specific settings
    :param from_: Producer doesn't support FromSection, defaults to None
    """

    app: ProducerAppValues = Field(
        default=...,
        description=describe_attr("app", __doc__),
    )
    from_: None = Field(
        default=None,
        alias="from",
        title="From",
        description=describe_attr("from_", __doc__),
    )

    @computed_field
    @cached_property
    def _cleaner(self) -> ProducerAppCleaner:
        return ProducerAppCleaner(
            config=self.config,
            handlers=self.handlers,
            **self.model_dump(by_alias=True, exclude={"_cleaner", "from_", "to"}),
        )

    @override
    def apply_to_outputs(self, name: str, topic: TopicConfig) -> None:
        match topic.type:
            case OutputTopicTypes.ERROR:
                msg = "Producer apps do not support error topics"
                raise ValueError(msg)
            case _:
                super().apply_to_outputs(name, topic)

    @property
    @override
    def output_topic(self) -> KafkaTopic | None:
        return self.app.streams.output_topic

    @property
    @override
    def extra_output_topics(self) -> dict[str, KafkaTopic]:
        return self.app.streams.extra_output_topics

    @override
    def set_output_topic(self, topic: KafkaTopic) -> None:
        self.app.streams.output_topic = topic

    @override
    def add_extra_output_topic(self, topic: KafkaTopic, role: str) -> None:
        self.app.streams.extra_output_topics[role] = topic

    @property
    @override
    def helm_chart(self) -> str:
        return f"{self.repo_config.repository_name}/{AppType.PRODUCER_APP.value}"

    @override
    async def clean(self, dry_run: bool) -> None:
        await self._cleaner.clean(dry_run)



from abc import ABC

from pydantic import Field

from kpops.component_handlers.helm_wrapper.model import HelmRepoConfig
from kpops.components.base_components.helm_app import HelmApp
from kpops.utils.docstring import describe_attr

STREAMS_BOOTSTRAP_HELM_REPO = HelmRepoConfig(
    repository_name="bakdata-streams-bootstrap",
    url="https://bakdata.github.io/streams-bootstrap/",
)
STREAMS_BOOTSTRAP_VERSION = "2.9.0"


class StreamsBootstrap(HelmApp, ABC):
    """Base for components with a streams-bootstrap Helm chart.

    :param repo_config: Configuration of the Helm chart repo to be used for
        deploying the component, defaults to streams-bootstrap Helm repo
    :param version: Helm chart version, defaults to "2.9.0"
    """

    repo_config: HelmRepoConfig = Field(
        default=STREAMS_BOOTSTRAP_HELM_REPO,
        description=describe_attr("repo_config", __doc__),
    )
    version: str | None = Field(
        default=STREAMS_BOOTSTRAP_VERSION,
        description=describe_attr("version", __doc__),
    )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from kpops.component_handlers.kafka_connect.connect_wrapper import ConnectWrapper
from kpops.component_handlers.kafka_connect.exception import (
    ConnectorNotFoundException,
    ConnectorStateException,
)
from kpops.utils.colorify import magentaify
from kpops.utils.dict_differ import render_diff

if TYPE_CHECKING:
    try:
        from typing import Self  # pyright: ignore[reportAttributeAccessIssue]
    except ImportError:
        from typing_extensions import Self

    from kpops.component_handlers.kafka_connect.model import KafkaConnectorConfig
    from kpops.config import KpopsConfig

log = logging.getLogger("KafkaConnectHandler")


class KafkaConnectHandler:
    def __init__(self, connect_wrapper: ConnectWrapper):
        self._connect_wrapper = connect_wrapper

    async def create_connector(
        self, connector_config: KafkaConnectorConfig, *, dry_run: bool
    ) -> None:
        """Create a connector.

        If the connector exists the config of that connector gets updated.

        :param connector_config: The connector config.
        :param dry_run: Whether the connector creation should be run in dry run mode.
        """
        if dry_run:
            await self.__dry_run_connector_creation(connector_config)
        else:
            try:
                await self._connect_wrapper.get_connector(connector_config.name)
                await self._connect_wrapper.update_connector_config(connector_config)

            except ConnectorNotFoundException:
                await self._connect_wrapper.create_connector(connector_config)

    async def destroy_connector(self, connector_name: str, *, dry_run: bool) -> None:
        """Delete a connector resource from the cluster.

        :param connector_name: The connector name.
        :param dry_run: Whether the connector deletion should be run in dry run mode.
        """
        if dry_run:
            await self.__dry_run_connector_deletion(connector_name)
        else:
            try:
                await self._connect_wrapper.get_connector(connector_name)
                await self._connect_wrapper.delete_connector(connector_name)

            except ConnectorNotFoundException:
                log.warning(
                    f"Connector Destruction: the connector {connector_name} does not exist. Skipping."
                )

    async def __dry_run_connector_creation(
        self, connector_config: KafkaConnectorConfig
    ) -> None:
        connector_name = connector_config.name
        try:
            connector = await self._connect_wrapper.get_connector(connector_name)

            log.info(f"Connector Creation: connector {connector_name} already exists.")
            if diff := render_diff(connector.config, connector_config.model_dump()):
                log.info(f"Updating config:\n{diff}")

            log.debug(connector_config.model_dump())
            log.debug(f"PUT /connectors/{connector_name}/config HTTP/1.1")
            log.debug(f"HOST: {self._connect_wrapper.url}")
        except ConnectorNotFoundException:
            diff = render_diff({}, connector_config.model_dump())
            log.info(
                f"Connector Creation: connector {connector_name} does not exist. Creating connector with config:\n{diff}"
            )
            log.debug("POST /connectors HTTP/1.1")
            log.debug(f"HOST: {self._connect_wrapper.url}")

        errors = await self._connect_wrapper.validate_connector_config(connector_config)
        if len(errors) > 0:
            formatted_errors = "\n".join(errors)
            msg = f"Connector Creation: validating the connector config for connector {connector_name} resulted in the following errors: {formatted_errors}"
            raise ConnectorStateException(msg)
        else:
            log.info(
                f"Connector Creation: connector config for {connector_name} is valid!"
            )

    async def __dry_run_connector_deletion(self, connector_name: str) -> None:
        try:
            await self._connect_wrapper.get_connector(connector_name)
            log.info(
                magentaify(
                    f"Connector Destruction: connector {connector_name} already exists. Deleting connector."
                )
            )
            log.debug(f"DELETE /connectors/{connector_name} HTTP/1.1")
            log.debug(f"HOST: {self._connect_wrapper.url}")
        except ConnectorNotFoundException:
            log.warning(
                f"Connector Destruction: connector {connector_name} does not exist and cannot be deleted. Skipping."
            )

    @classmethod
    def from_kpops_config(cls, config: KpopsConfig) -> Self:
        return cls(
            connect_wrapper=ConnectWrapper(config.kafka_connect),
        )



from __future__ import annotations

import logging
import re
from abc import ABC

from pydantic import ConfigDict, Field
from typing_extensions import override

from kpops.components.base_components.pipeline_component import PipelineComponent
from kpops.utils.docstring import describe_attr
from kpops.utils.pydantic import CamelCaseConfigModel, DescConfigModel

log = logging.getLogger("KubernetesApp")

KUBERNETES_NAME_CHECK_PATTERN = re.compile(
    r"^(?![0-9]+$)(?!.*-$)(?!-)[a-z0-9-.]{1,253}(?<!_)$"
)


class KubernetesAppValues(CamelCaseConfigModel, DescConfigModel):
    """Settings specific to Kubernetes apps."""

    model_config = ConfigDict(
        extra="allow",
    )


class KubernetesApp(PipelineComponent, ABC):
    """Base class for all Kubernetes apps.

    All built-in components are Kubernetes apps, except for the Kafka connectors.

    :param namespace: Kubernetes namespace in which the component shall be deployed
    :param app: Application-specific settings
    """

    namespace: str = Field(
        default=...,
        description=describe_attr("namespace", __doc__),
    )
    app: KubernetesAppValues = Field(
        default=...,
        description=describe_attr("app", __doc__),
    )

    @override
    def _validate_custom(self) -> None:
        super()._validate_custom()
        self.validate_kubernetes_name(self.name)

    @staticmethod
    def validate_kubernetes_name(name: str) -> None:
        """Check if a name is valid for a Kubernetes resource.

        :param name: Name that is to be used for the resource
        :raises ValueError: The component name {name} is invalid for Kubernetes.
        """
        if not bool(KUBERNETES_NAME_CHECK_PATTERN.match(name)):
            msg = f"The component name {name} is invalid for Kubernetes."
            raise ValueError(msg)



from __future__ import annotations

from pathlib import Path

from pydantic import AnyHttpUrl, Field, TypeAdapter
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)
from typing_extensions import override

from kpops.component_handlers.helm_wrapper.model import HelmConfig, HelmDiffConfig
from kpops.utils.docstring import describe_object
from kpops.utils.pydantic import YamlConfigSettingsSource

ENV_PREFIX = "KPOPS_"


class TopicNameConfig(BaseSettings):
    """Configure the topic name variables you can use in the pipeline definition."""

    default_output_topic_name: str = Field(
        default="${pipeline.name}-${component.name}",
        description="Configures the value for the variable ${output_topic_name}",
    )
    default_error_topic_name: str = Field(
        default="${pipeline.name}-${component.name}-error",
        description="Configures the value for the variable ${error_topic_name}",
    )


class SchemaRegistryConfig(BaseSettings):
    """Configuration for Schema Registry."""

    enabled: bool = Field(
        default=False,
        description="Whether the Schema Registry handler should be initialized.",
    )
    url: AnyHttpUrl = Field(
        default=TypeAdapter(AnyHttpUrl).validate_python("http://localhost:8081"),  # pyright: ignore[reportCallIssue]
        description="Address of the Schema Registry.",
    )
    timeout: int | float = Field(
        default=30, description="Operation timeout in seconds."
    )


class KafkaRestConfig(BaseSettings):
    """Configuration for Kafka REST Proxy."""

    url: AnyHttpUrl = Field(
        default=TypeAdapter(AnyHttpUrl).validate_python("http://localhost:8082"),  # pyright: ignore[reportCallIssue]
        description="Address of the Kafka REST Proxy.",
    )
    timeout: int | float = Field(
        default=30, description="Operation timeout in seconds."
    )


class KafkaConnectConfig(BaseSettings):
    """Configuration for Kafka Connect."""

    url: AnyHttpUrl = Field(
        default=TypeAdapter(AnyHttpUrl).validate_python("http://localhost:8083"),  # pyright: ignore[reportCallIssue]
        description="Address of Kafka Connect.",
    )
    timeout: int | float = Field(
        default=30, description="Operation timeout in seconds."
    )


class KpopsConfig(BaseSettings):
    """Global configuration for KPOps project."""

    components_module: str | None = Field(
        default=None,
        description="Custom Python module defining project-specific KPOps components",
    )
    pipeline_base_dir: Path = Field(
        default=Path(),
        description="Base directory to the pipelines (default is current working directory)",
    )
    kafka_brokers: str = Field(
        default=...,
        examples=[
            "broker1:9092,broker2:9092,broker3:9092",
        ],
        description="The comma separated Kafka brokers address.",
    )
    defaults_filename_prefix: str = Field(
        default="defaults",
        description="The name of the defaults file and the prefix of the defaults environment file.",
    )
    topic_name_config: TopicNameConfig = Field(
        default=TopicNameConfig(),
        description=describe_object(TopicNameConfig.__doc__),
    )
    schema_registry: SchemaRegistryConfig = Field(
        default=SchemaRegistryConfig(),
        description=describe_object(SchemaRegistryConfig.__doc__),
    )
    kafka_rest: KafkaRestConfig = Field(
        default=KafkaRestConfig(),
        description=describe_object(KafkaRestConfig.__doc__),
    )
    kafka_connect: KafkaConnectConfig = Field(
        default=KafkaConnectConfig(),
        description=describe_object(KafkaConnectConfig.__doc__),
    )
    create_namespace: bool = Field(
        default=False,
        description="Flag for `helm upgrade --install`. Create the release namespace if not present.",
    )
    helm_config: HelmConfig = Field(
        default=HelmConfig(),
        description="Global flags for Helm.",
    )
    helm_diff_config: HelmDiffConfig = Field(
        default=HelmDiffConfig(),
        description="Configure Helm Diff.",
    )
    retain_clean_jobs: bool = Field(
        default=False,
        description="Whether to retain clean up jobs in the cluster or uninstall the, after completion.",
    )

    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX, env_nested_delimiter="__")

    @override
    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ):
        """Change the precedence of sources that the Pydantic settings are compiled from.

        Environment variables (`env_settings`) take precedence over command line flags (`init_settings`).
        Command line flags (`init_settings`) take precedence over settings from a
        config.yaml (`YamlConfigSettingsSource(settings_cls)`) and so forth.
        """
        return (
            env_settings,
            init_settings,
            YamlConfigSettingsSource(settings_cls),
            dotenv_settings,
            file_secret_settings,
        )



import asyncio
import logging
from asyncio import TimeoutError
from collections.abc import Coroutine
from typing import Any, TypeVar

log = logging.getLogger("Timeout")

T = TypeVar("T")


async def timeout(coro: Coroutine[Any, Any, T], *, secs: int = 0) -> T | None:
    """Set a timeout for a given lambda function.

    :param coro: The callable function
    :param secs: The timeout in seconds.
    """
    try:
        task = asyncio.create_task(coro)
        if secs == 0:
            return await task
        else:
            return await asyncio.wait_for(task, timeout=secs)
    except TimeoutError:
        log.exception(
            f"Kafka Connect operation {coro.__name__} timed out after {secs} seconds. To increase the duration, set the `timeout` option in config.yaml."
        )



from logging import Logger

from kpops.component_handlers.helm_wrapper.helm import Helm
from kpops.component_handlers.helm_wrapper.helm_diff import HelmDiff


class DryRunHandler:
    def __init__(self, helm: Helm, helm_diff: HelmDiff, namespace: str) -> None:
        self._helm = helm
        self._helm_diff = helm_diff
        self.namespace = namespace

    def print_helm_diff(self, stdout: str, helm_release_name: str, log: Logger) -> None:
        """Print the diff of the last and current release of this component.

        :param stdout: The output of a Helm command that installs or upgrades the release
        :param helm_release_name: The Helm release name
        :param log: The Logger object of the component class
        """
        current_release = list(
            self._helm.get_manifest(helm_release_name, self.namespace)
        )
        if current_release:
            log.info(f"Helm release {helm_release_name} already exists")
        else:
            log.info(f"Helm release {helm_release_name} does not exist")
        new_release = Helm.load_manifest(stdout)
        self._helm_diff.log_helm_diff(log, current_release, new_release)



from __future__ import annotations

import asyncio
import logging
import re
import subprocess
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING

import yaml

from kpops.component_handlers.helm_wrapper.exception import ReleaseNotFoundException
from kpops.component_handlers.helm_wrapper.model import (
    HelmChart,
    HelmConfig,
    HelmTemplate,
    HelmTemplateFlags,
    HelmUpgradeInstallFlags,
    RepoAuthFlags,
    Version,
)
from kpops.component_handlers.kubernetes.model import KubernetesManifest
from kpops.components.base_components.models.resource import Resource

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

log = logging.getLogger("Helm")


class Helm:
    def __init__(self, helm_config: HelmConfig) -> None:
        self._context = helm_config.context
        self._debug = helm_config.debug
        self._version = self.get_version()
        if self._version.major != 3:
            msg = f"The supported Helm version is 3.x.x. The current Helm version is {self._version.major}.{self._version.minor}.{self._version.patch}"
            raise RuntimeError(msg)

    def add_repo(
        self,
        repository_name: str,
        repository_url: str,
        repo_auth_flags: RepoAuthFlags | None = None,
    ) -> None:
        if repo_auth_flags is None:
            repo_auth_flags = RepoAuthFlags()
        command = [
            "helm",
            "repo",
            "add",
            repository_name,
            repository_url,
        ]
        command.extend(repo_auth_flags.to_command())

        try:
            self.__execute(command)
        except (ReleaseNotFoundException, RuntimeError) as e:
            if (
                len(e.args) == 1
                and re.match(
                    "Error: repository name (.*) already exists, please specify a different name",
                    e.args[0],
                )
                is not None
            ):
                log.exception(f"Could not add repository {repository_name}.")
            else:
                raise

        if self._version.minor > 7:
            self.__execute(["helm", "repo", "update", repository_name])
        else:
            self.__execute(["helm", "repo", "update"])

    async def upgrade_install(
        self,
        release_name: str,
        chart: str,
        dry_run: bool,
        namespace: str,
        values: dict,
        flags: HelmUpgradeInstallFlags | None = None,
    ) -> str:
        """Prepare and execute the `helm upgrade --install` command."""
        if flags is None:
            flags = HelmUpgradeInstallFlags()
        with tempfile.NamedTemporaryFile("w", delete=False) as values_file:
            yaml.safe_dump(values, values_file)

            command = [
                "helm",
                "upgrade",
                release_name,
                chart,
                "--install",
                "--namespace",
                namespace,
                "--values",
                values_file.name,
            ]
            command.extend(flags.to_command())
            if dry_run:
                command.append("--dry-run")
            return await self.__async_execute(command)

    async def uninstall(
        self,
        namespace: str,
        release_name: str,
        dry_run: bool,
    ) -> str | None:
        """Prepare and execute the `helm uninstall` command."""
        command = [
            "helm",
            "uninstall",
            release_name,
            "--namespace",
            namespace,
        ]
        if dry_run:
            command.append("--dry-run")
        try:
            return await self.__async_execute(command)
        except ReleaseNotFoundException:
            log.warning(
                f"Release with name {release_name} not found. Could not uninstall app."
            )

    def template(
        self,
        release_name: str,
        chart: str,
        namespace: str,
        values: dict,
        flags: HelmTemplateFlags | None = None,
    ) -> Resource:
        """From Helm: Render chart templates locally and display the output.

        Any values that would normally be looked up or retrieved in-cluster will
        be faked locally. Additionally, none of the server-side testing of chart
        validity (e.g. whether an API is supported) is done.

        :param str release_name: the release name for which the command is ran
        :param chart: Helm chart to be templated
        :param namespace: The Kubernetes namespace the command should execute in
        :param values: `values.yaml` to be used
        :param flags: the flags to be set for `helm template`, defaults to HelmTemplateFlags()
        :return: the rendered resource (list of Kubernetes manifests)
        """
        if flags is None:
            flags = HelmTemplateFlags()
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as values_file:
            yaml.safe_dump(values, values_file)
            command = [
                "helm",
                "template",
                release_name,
                chart,
                "--namespace",
                namespace,
                "--values",
                values_file.name,
            ]
            command.extend(flags.to_command())
            output = self.__execute(command)
            manifests = KubernetesManifest.from_yaml(output)
            return list(manifests)

    def get_manifest(self, release_name: str, namespace: str) -> Iterable[HelmTemplate]:
        command = [
            "helm",
            "get",
            "manifest",
            release_name,
            "--namespace",
            namespace,
        ]

        try:
            stdout = self.__execute(command=command)
            return Helm.load_manifest(stdout)
        except ReleaseNotFoundException:
            return ()

    def get_version(self) -> Version:
        command = ["helm", "version", "--short"]
        short_version = self.__execute(command)
        version_match = re.search(r"^v(\d+(?:\.\d+){0,2})", short_version)
        if version_match is None:
            msg = f"Could not parse the Helm version.\n\nHelm output:\n{short_version}"
            raise RuntimeError(msg)
        version = map(int, version_match.group(1).split("."))
        return Version(*version)

    @staticmethod
    def load_manifest(yaml_contents: str) -> Iterator[HelmTemplate]:
        is_beginning: bool = False
        template_name = None
        current_yaml_doc: list[str] = []
        for line in HelmChart(yaml_contents):
            if line.startswith("---"):
                is_beginning = True
                if template_name and current_yaml_doc:
                    manifests = KubernetesManifest.from_yaml(
                        "\n".join(current_yaml_doc)
                    )
                    manifest = next(manifests)  # only 1 manifest
                    yield HelmTemplate(Path(template_name), manifest)
                    template_name = None
                    current_yaml_doc.clear()
            elif is_beginning:
                template_name = HelmTemplate.parse_source(line)
                is_beginning = False
            else:
                current_yaml_doc.append(line)

    def __execute(self, command: list[str]) -> str:
        command = self.__set_global_flags(command)
        log.debug(f"Executing {' '.join(command)}")
        process = subprocess.run(
            command,
            check=False,
            capture_output=True,
            text=True,
        )
        Helm.parse_helm_command_stderr_output(process.stderr)
        log.debug(process.stdout)
        return process.stdout

    async def __async_execute(self, command: list[str]):
        command = self.__set_global_flags(command)
        log.debug(f"Executing {' '.join(command)}")
        proc = await asyncio.create_subprocess_exec(
            *command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        stdout, stderr = await proc.communicate()
        Helm.parse_helm_command_stderr_output(stderr.decode())
        log.debug(stdout)
        return stdout.decode()

    def __set_global_flags(self, command: list[str]) -> list[str]:
        if self._context:
            log.debug(f"Changing the Kubernetes context to {self._context}")
            command.extend(["--kube-context", self._context])
        if self._debug:
            log.debug("Enabling verbose mode.")
            command.append("--debug")
        return command

    @staticmethod
    def parse_helm_command_stderr_output(stderr: str) -> None:
        for line in stderr.splitlines():
            lower = line.lower()
            if "release: not found" in lower:
                raise ReleaseNotFoundException
            elif "error" in lower:
                raise RuntimeError(stderr)
            elif "warning" in lower:
                log.warning(line)



from __future__ import annotations

import json
from collections import UserDict
from collections.abc import Iterator

import yaml

from kpops.utils.types import JsonType

K8S_LABEL_MAX_LEN = 63


class KubernetesManifest(UserDict[str, JsonType]):
    """Representation of a Kubernetes API object as YAML/JSON mapping."""

    @classmethod
    def from_yaml(
        cls, /, content: str
    ) -> Iterator[KubernetesManifest]:  # TODO: typing.Self for Python 3.11+
        manifests: Iterator[dict[str, JsonType]] = yaml.load_all(content, yaml.Loader)
        for manifest in manifests:
            yield cls(manifest)

    @classmethod
    def from_json(
        cls, /, content: str
    ) -> KubernetesManifest:  # TODO: typing.Self for Python 3.11+
        manifest: dict[str, JsonType] = json.loads(content)
        return cls(manifest)



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from difflib import Differ
from enum import Enum
from typing import TYPE_CHECKING, Generic, TypeVar

import typer
import yaml
from dictdiffer import diff, patch

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence

differ = Differ()


class DiffType(str, Enum):
    ADD = "add"
    CHANGE = "change"
    REMOVE = "remove"

    @staticmethod
    def from_str(label: str) -> DiffType:
        return DiffType[label.upper()]


T = TypeVar("T")


@dataclass
class Change(Generic[T]):  # Generic NamedTuple requires Python 3.11+
    old_value: T
    new_value: T

    @staticmethod
    def factory(type: DiffType, change: T | tuple[T, T]) -> Change:
        match type:
            case DiffType.ADD:
                return Change(None, change)
            case DiffType.REMOVE:
                return Change(change, None)
            case DiffType.CHANGE if isinstance(change, tuple):
                return Change(*change)
        msg = f"{type} is not part of {DiffType}"
        raise ValueError(msg)


@dataclass
class Diff(Generic[T]):
    diff_type: DiffType
    key: str
    change: Change[T]

    @staticmethod
    def from_dicts(
        d1: dict, d2: dict, ignore: set[str] | None = None
    ) -> Iterator[Diff]:
        for diff_type, keys, changes in diff(d1, d2, ignore=ignore):
            if not isinstance(changes_tmp := changes, list):
                changes_tmp = [("", changes)]
            for key, change in changes_tmp:
                yield Diff(
                    DiffType.from_str(diff_type),
                    Diff.__find_changed_key(keys, key),
                    Change.factory(diff_type, change),
                )

    @staticmethod
    def __find_changed_key(key_1: list[str] | str, key_2: str = "") -> str:
        """Generate a string that points to the changed key in the dictionary."""
        if isinstance(key_1, list) and len(key_1) > 1:
            return f"{key_1[0]}[{key_1[1]}]"
        if not key_1:
            return key_2
        if not key_2:
            return "".join(key_1)
        return f"{key_1}.{key_2}"


def render_diff(d1: Mapping, d2: Mapping, ignore: set[str] | None = None) -> str | None:
    differences = list(diff(d1, d2, ignore=ignore))
    if not differences:
        return None

    d2_filtered: Mapping = patch(differences, d1)
    return "".join(
        colorize_diff(
            differ.compare(
                to_yaml(d1) if d1 else "",
                to_yaml(d2_filtered) if d2_filtered else "",
            )
        )
    )


def colorize_diff(input: Iterable[str]) -> Iterator[str]:
    for line in input:
        yield colorize_line(line)


def colorize_line(line: str) -> str:
    if line.startswith("-"):
        return typer.style(line, fg=typer.colors.RED)
    if line.startswith("+"):
        return typer.style(line, fg=typer.colors.GREEN)
    if line.startswith("?"):
        return typer.style(line, fg=typer.colors.YELLOW)
    return line


def to_yaml(data: Mapping) -> Sequence[str]:
    return yaml.dump(data, sort_keys=True).splitlines(keepends=True)



from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import Path

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import override

from kpops.component_handlers.helm_wrapper.exception import ParseError
from kpops.component_handlers.kubernetes.model import KubernetesManifest
from kpops.utils.docstring import describe_attr
from kpops.utils.pydantic import DescConfigModel


class HelmDiffConfig(BaseModel):
    ignore: set[str] = Field(
        default_factory=set,
        description="Set of keys that should not be checked.",
        examples=["- name\n- imageTag"],
    )


class RepoAuthFlags(DescConfigModel):
    """Authorisation-related flags for `helm repo`.

    :param username: Username, defaults to None
    :param password: Password, defaults to None
    :param ca_file: Path to CA bundle file to verify certificates of HTTPS-enabled servers, defaults to None
    :param cert_file: Path to SSL certificate file to identify HTTPS client, defaults to None
    :param insecure_skip_tls_verify: If true, Kubernetes API server's certificate will not be checked for validity
        , defaults to False
    """

    username: str | None = Field(
        default=None, description=describe_attr("username", __doc__)
    )
    password: str | None = Field(
        default=None, description=describe_attr("password", __doc__)
    )
    ca_file: Path | None = Field(
        default=None, description=describe_attr("ca_file", __doc__)
    )
    cert_file: Path | None = Field(
        default=None, description=describe_attr("cert_file", __doc__)
    )
    insecure_skip_tls_verify: bool = Field(
        default=False, description=describe_attr("insecure_skip_tls_verify", __doc__)
    )

    def to_command(self) -> list[str]:
        command: list[str] = []
        if self.username:
            command.extend(["--username", self.username])
        if self.password:
            command.extend(["--password", self.password])
        if self.ca_file:
            command.extend(["--ca-file", str(self.ca_file)])
        if self.cert_file:
            command.extend(["--cert-file", str(self.cert_file)])
        if self.insecure_skip_tls_verify:
            command.append("--insecure-skip-tls-verify")
        return command


class HelmRepoConfig(DescConfigModel):
    """Helm repository configuration.

    :param repository_name: Name of the Helm repository
    :param url: URL to the Helm repository
    :param repo_auth_flags: Authorisation-related flags
    """

    repository_name: str = Field(
        default=..., description=describe_attr("repository_name", __doc__)
    )
    url: str = Field(default=..., description=describe_attr("url", __doc__))
    repo_auth_flags: RepoAuthFlags = Field(
        default=RepoAuthFlags(), description=describe_attr("repo_auth_flags", __doc__)
    )


class HelmConfig(DescConfigModel):
    """Global Helm configuration.

    :param context: Name of kubeconfig context (`--kube-context`)
    :param debug: Run Helm in Debug mode
    :param api_version: Kubernetes API version used for `Capabilities.APIVersions`
    """

    context: str | None = Field(
        default=None,
        description=describe_attr("context", __doc__),
        examples=["dev-storage"],
    )
    debug: bool = Field(
        default=False,
        description=describe_attr("debug", __doc__),
    )
    api_version: str | None = Field(
        default=None,
        title="API version",
        description=describe_attr("api_version", __doc__),
    )


class HelmFlags(RepoAuthFlags):
    set_file: dict[str, Path] = Field(default_factory=dict)
    create_namespace: bool = False
    version: str | None = None
    force: bool = False
    timeout: str = "5m0s"
    wait: bool = True
    wait_for_jobs: bool = False

    model_config = ConfigDict(
        extra="allow",
    )

    @override
    def to_command(self) -> list[str]:
        command = super().to_command()
        if self.set_file:
            command.extend(
                [
                    "--set-file",
                    ",".join([f"{key}={path}" for key, path in self.set_file.items()]),
                ]
            )
        if self.create_namespace:
            command.append("--create-namespace")
        if self.version:
            command.extend(["--version", self.version])
        if self.force:
            command.append("--force")
        if self.timeout:
            command.extend(["--timeout", self.timeout])
        if self.wait:
            command.append("--wait")
        if self.wait_for_jobs:
            command.append("--wait-for-jobs")
        return command


class HelmUpgradeInstallFlags(HelmFlags): ...


class HelmTemplateFlags(HelmFlags):
    api_version: str | None = None

    @override
    def to_command(self) -> list[str]:
        command = super().to_command()
        if self.api_version:
            command.extend(["--api-versions", self.api_version])
        return command


HELM_SOURCE_PREFIX = "# Source: "


@dataclass
class HelmTemplate:
    filepath: Path
    manifest: KubernetesManifest

    @staticmethod
    def parse_source(source: str) -> str:
        """Parse source path from comment at the beginning of the YAML doc.

        :Example:
        .. code-block:: yaml
            # Source: chart/templates/serviceaccount.yaml
        """
        if not source.startswith(HELM_SOURCE_PREFIX):
            msg = "Not a valid Helm template source"
            raise ParseError(msg)
        return source.removeprefix(HELM_SOURCE_PREFIX).strip()


# Indicates the beginning of `NOTES:` section in the output of `helm install` or
# `helm upgrade`
HELM_NOTES = "\n\nNOTES:\n"
HELM_MANIFEST = "MANIFEST:\n"


@dataclass(frozen=True)
class HelmChart:
    content: str

    def __iter__(self) -> Iterator[str]:
        yield from self.manifest.splitlines()
        yield "---"  # add final divider to make parsing easier

    @property
    def manifest(self) -> str:
        """Reads the manifest section of Helm stdout.

        `helm upgrade --install` output message contains three sections in the following order:

        - HOOKS
        - MANIFEST
        - NOTES (optional)

        The content of the manifest is used to create the diff. If a NOTES.txt exists in the Helm chart, the NOTES
        section will be included in the output.

        It is important to note that the `helm get manifest` command only returns the manifests without the MANIFEST
        header in the stdout. Instead, the output starts with `---`.

        :return: The content of the manifest section
        """
        manifest_start = (
            self.content.index(HELM_MANIFEST) + len(HELM_MANIFEST)
            if HELM_MANIFEST in self.content
            else self.content.index("---")
        )

        manifest_end = (
            self.content.index(HELM_NOTES) if HELM_NOTES in self.content else -1
        )

        return self.content[manifest_start:manifest_end]


@dataclass(frozen=True)
class Version:
    major: int
    minor: int = 0
    patch: int = 0



from pydantic import ConfigDict, Field

from kpops.components.base_components.kafka_app import (
    KafkaAppValues,
    KafkaStreamsConfig,
)
from kpops.utils.docstring import describe_attr


class ProducerStreamsConfig(KafkaStreamsConfig):
    """Kafka Streams settings specific to Producer."""


class ProducerAppValues(KafkaAppValues):
    """Settings specific to producers.

    :param streams: Kafka Streams settings
    """

    streams: ProducerStreamsConfig = Field(
        default=..., description=describe_attr("streams", __doc__)
    )

    model_config = ConfigDict(extra="allow")



import hashlib
import logging

log = logging.getLogger("K8sUtils")


def trim(max_len: int, name: str, suffix: str) -> str:
    """Shortens long K8s identifiers.

    Kubernetes has a character limit for certain identifiers.
    If the name exceeds the character limit:
    1. trim the string and fetch the first ``max_len - len(suffix)`` characters.
    2. replace the last 6 characters with the SHA-1 encoded string (with "-") to avoid collision
    3. append the suffix if given

    :param name: The name to be shortened.
    :param suffix: The release suffix to preserve
    :return: Trimmed + hashed version of the name if it exceeds the character length otherwise the actual name
    """
    if len(name) > max_len:
        exact_name = name[: max_len - len(suffix)]
        hash_name = hashlib.sha1(name.encode()).hexdigest()
        new_name = exact_name[:-6] + "-" + hash_name[:5] + suffix
        log.critical(
            f"Kubernetes identifier '{name}' exceeds character limit. Truncating and hashing to {max_len} characters: \n {name} --> {new_name}"
        )
        return new_name
    return name



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING

import httpx

from kpops.component_handlers.kafka_connect.exception import (
    ConnectorNotFoundException,
    KafkaConnectError,
)
from kpops.component_handlers.kafka_connect.model import (
    KafkaConnectConfigErrorResponse,
    KafkaConnectorConfig,
    KafkaConnectResponse,
)

if TYPE_CHECKING:
    from pydantic import AnyHttpUrl

    from kpops.config import KafkaConnectConfig

HEADERS = {"Accept": "application/json", "Content-Type": "application/json"}

log = logging.getLogger("KafkaConnectAPI")


class ConnectWrapper:
    """Wraps Kafka Connect APIs."""

    def __init__(self, config: KafkaConnectConfig) -> None:
        self._config: KafkaConnectConfig = config
        self._client = httpx.AsyncClient(timeout=config.timeout)

    @property
    def url(self) -> AnyHttpUrl:
        return self._config.url

    async def create_connector(
        self, connector_config: KafkaConnectorConfig
    ) -> KafkaConnectResponse:
        """Create a new connector.

        API Reference: https://docs.confluent.io/platform/current/connect/references/restapi.html#post--connectors
        :param connector_config: The config of the connector
        :return: The current connector info if successful.
        """
        config_json = connector_config.model_dump()
        connect_data = {"name": connector_config.name, "config": config_json}
        response = await self._client.post(
            url=f"{self.url}connectors", headers=HEADERS, json=connect_data
        )
        if response.status_code == httpx.codes.CREATED:
            log.info(f"Connector {connector_config.name} created.")
            log.debug(response.json())
            return KafkaConnectResponse(**response.json())
        elif response.status_code == httpx.codes.CONFLICT:
            log.warning(
                "Rebalancing in progress while creating a connector... Retrying..."
            )

            await asyncio.sleep(1)
            await self.create_connector(connector_config)

        raise KafkaConnectError(response)

    async def get_connector(self, connector_name: str | None) -> KafkaConnectResponse:
        """Get information about the connector.

        API Reference: https://docs.confluent.io/platform/current/connect/references/restapi.html#get--connectors-(string-name)
        :param connector_name: Nameof the crated connector
        :return: Information about the connector.
        """
        if connector_name is None:
            msg = "Connector name not set"
            raise Exception(msg)
        response = await self._client.get(
            url=f"{self.url}connectors/{connector_name}", headers=HEADERS
        )
        if response.status_code == httpx.codes.OK:
            log.info(f"Connector {connector_name} exists.")
            log.debug(response.json())
            return KafkaConnectResponse(**response.json())
        elif response.status_code == httpx.codes.NOT_FOUND:
            log.info(f"The named connector {connector_name} does not exists.")
            raise ConnectorNotFoundException
        elif response.status_code == httpx.codes.CONFLICT:
            log.warning(
                "Rebalancing in progress while getting a connector... Retrying..."
            )
            await asyncio.sleep(1)
            await self.get_connector(connector_name)
        raise KafkaConnectError(response)

    async def update_connector_config(
        self, connector_config: KafkaConnectorConfig
    ) -> KafkaConnectResponse:
        """Create or update a connector.

        Create a new connector using the given configuration, or update the
        configuration for an existing connector.
        :param connector_config: Configuration parameters for the connector.
        :return: Information about the connector after the change has been made.
        """
        connector_name = connector_config.name

        config_json = connector_config.model_dump()
        response = await self._client.put(
            url=f"{self.url}connectors/{connector_name}/config",
            headers=HEADERS,
            json=config_json,
        )

        data: dict = response.json()
        if response.status_code == httpx.codes.OK:
            log.info(f"Config for connector {connector_name} updated.")
            log.debug(data)
            return KafkaConnectResponse(**data)
        if response.status_code == httpx.codes.CREATED:
            log.info(f"Connector {connector_name} created.")
            log.debug(data)
            return KafkaConnectResponse(**data)
        elif response.status_code == httpx.codes.CONFLICT:
            log.warning(
                "Rebalancing in progress while updating a connector... Retrying..."
            )
            await asyncio.sleep(1)
            await self.update_connector_config(connector_config)
        raise KafkaConnectError(response)

    async def validate_connector_config(
        self, connector_config: KafkaConnectorConfig
    ) -> list[str]:
        """Validate connector config using the given configuration.

        :param connector_config: Configuration parameters for the connector.
        :raises KafkaConnectError: Kafka Konnect error
        :return: List of all found errors
        """
        response = await self._client.put(
            url=f"{self.url}connector-plugins/{connector_config.class_name}/config/validate",
            headers=HEADERS,
            json=connector_config.model_dump(),
        )

        if response.status_code == httpx.codes.OK:
            kafka_connect_error_response = KafkaConnectConfigErrorResponse(
                **response.json()
            )

            errors: list[str] = []
            if kafka_connect_error_response.error_count > 0:
                for config in kafka_connect_error_response.configs:
                    if len(config.value.errors) > 0:
                        for error in config.value.errors:
                            errors.append(
                                f"Found error for field {config.value.name}: {error}"
                            )
            return errors
        raise KafkaConnectError(response)

    async def delete_connector(self, connector_name: str) -> None:
        """Delete a connector, halting all tasks and deleting its configuration.

        API Reference:
            https://docs.confluent.io/platform/current/connect/references/restapi.html#delete--connectors-(string-name)-.
        :param connector_name: Configuration parameters for the connector.
        :raises ConnectorNotFoundException: Connector not found
        """
        response = await self._client.delete(
            url=f"{self.url}connectors/{connector_name}", headers=HEADERS
        )
        if response.status_code == httpx.codes.NO_CONTENT:
            log.info(f"Connector {connector_name} deleted.")
            return
        elif response.status_code == httpx.codes.NOT_FOUND:
            log.info(f"The named connector {connector_name} does not exists.")
            raise ConnectorNotFoundException
        elif response.status_code == httpx.codes.CONFLICT:
            log.warning(
                "Rebalancing in progress while deleting a connector... Retrying..."
            )
            await asyncio.sleep(1)
            await self.delete_connector(connector_name)
        raise KafkaConnectError(response)



from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import TypeAlias

# JSON values
JsonType: TypeAlias = (
    Mapping[str, "JsonType"] | Sequence["JsonType"] | str | int | float | bool | None
)



from collections.abc import Mapping, Sequence
from typing import Any, TypeAlias

# representation of final resource for component, e.g. a list of Kubernetes manifests
Resource: TypeAlias = Sequence[Mapping[str, Any]]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, TypeAlias

from schema_registry.client.schema import AvroSchema, JsonSchema

if TYPE_CHECKING:
    from kpops.components.base_components.models import ModelName, ModelVersion

Schema: TypeAlias = AvroSchema | JsonSchema


class SchemaProvider(ABC):
    @abstractmethod
    def provide_schema(
        self, schema_class: str, models: dict[ModelName, ModelVersion]
    ) -> Schema: ...



from __future__ import annotations

from pydantic import ConfigDict, Field

from kpops.components.base_components.models import ModelName, ModelVersion, TopicName
from kpops.components.base_components.models.topic import KafkaTopic, TopicConfig
from kpops.utils.docstring import describe_attr
from kpops.utils.pydantic import DescConfigModel


class ToSection(DescConfigModel):
    """Holds multiple output topics.

    :param topics: Output topics
    :param models: Data models
    """

    topics: dict[TopicName, TopicConfig] = Field(
        default={}, description=describe_attr("topics", __doc__)
    )
    models: dict[ModelName, ModelVersion] = Field(
        default={}, description=describe_attr("models", __doc__)
    )

    model_config = ConfigDict(extra="forbid")

    @property
    def kafka_topics(self) -> list[KafkaTopic]:
        return [
            KafkaTopic(name=topic_name, config=topic_config)
            for topic_name, topic_config in self.topics.items()
        ]



from kpops.component_handlers.utils.exception import HttpxException


class ConnectorNotFoundException(Exception):
    pass


class ConnectorStateException(Exception):
    pass


class KafkaConnectError(HttpxException):
    pass



import logging
from pathlib import Path
from typing import Any

import humps
from pydantic import BaseModel, ConfigDict, Field
from pydantic.fields import FieldInfo
from pydantic_settings import PydanticBaseSettingsSource
from typing_extensions import TypeVar, override

from kpops.utils.dict_ops import update_nested_pair
from kpops.utils.docstring import describe_object
from kpops.utils.yaml import load_yaml_file


def to_camel(s: str) -> str:
    """Convert snake_case to camelCase."""
    return humps.camelize(s)


def to_dash(s: str) -> str:
    """Convert PascalCase to dash-case."""
    return humps.depascalize(s).lower().replace("_", "-")


def to_snake(s: str) -> str:
    """Convert PascalCase to snake_case."""
    return humps.depascalize(s).lower()


def to_dot(s: str) -> str:
    """Convert snake_case to dot.notation."""
    return s.replace("_", ".")


def by_alias(model: BaseModel, field_name: str) -> str:
    """Return field alias if exists else field name.

    :param field_name: Name of the field to get alias of
    :param model: Model that owns the field
    """
    return model.model_fields.get(field_name, Field()).alias or field_name


_V = TypeVar("_V")


def exclude_by_value(
    dumped_model: dict[str, _V], *excluded_values: Any
) -> dict[str, _V]:
    """Strip all key-value pairs with certain values.

    :param dumped_model: Dumped model
    :param excluded_values: Excluded field values
    :return: Dumped model without excluded fields
    """
    return {
        field_name: field_value
        for field_name, field_value in dumped_model.items()
        if field_value not in excluded_values
    }


def exclude_by_name(
    dumped_model: dict[str, _V], *excluded_fields: str
) -> dict[str, _V]:
    """Strip all key-value pairs with certain field names.

    :param dumped_model: Dumped model
    :param excluded_fields: Excluded field names
    :return: Dumped model without excluded fields
    """
    return {
        field_name: field_value
        for field_name, field_value in dumped_model.items()
        if field_name not in excluded_fields
    }


def exclude_defaults(model: BaseModel, dumped_model: dict[str, _V]) -> dict[str, _V]:
    """Strip all key-value pairs with default values.

    :param model: Model
    :param dumped_model: Dumped model
    :return: Dumped model without defaults
    """
    default_fields = {
        field_name: field_info.default
        for field_name, field_info in model.model_fields.items()
    }
    return {
        field_name: field_value
        for field_name, field_value in dumped_model.items()
        if field_value
        not in (
            default_fields.get(field_name),
            default_fields.get(to_snake(field_name)),
        )
    }


def collect_fields(model: type[BaseModel]) -> dict[str, Any]:
    """Collect and return a ``dict`` of all fields in a settings class.

    :param model: settings class
    :return: ``dict`` of all fields in a settings class
    """
    seen_fields = {}
    for field_name, field_value in model.model_fields.items():
        if field_value.annotation and issubclass_patched(field_value.annotation):
            seen_fields[field_name] = collect_fields(field_value.annotation)
        else:
            seen_fields[field_name] = field_value
    return seen_fields


def issubclass_patched(
    __cls: type, __class_or_tuple: type | tuple[type, ...] = BaseModel
) -> bool:
    """Pydantic breaks ``issubclass``.

    .. code-block:: python
        issubclass(set[str], set)  # True
        issubclass(BaseSettings, BaseModel)  # True
        issubclass(set[str], BaseModel)  # raises Exception

    :param __cls: class to check
    :param __class_or_tuple: class(es) to check against, defaults to ``BaseModel``
    :return: Whether 'cls' is derived from another class or is the same class.
    """
    try:
        return issubclass(__cls, __class_or_tuple)
    except TypeError as e:
        if str(e) == "issubclass() arg 1 must be a class":
            return False
        raise


class CamelCaseConfigModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
    )


class DescConfigModel(BaseModel):
    @staticmethod
    def json_schema_extra(schema: dict[str, Any], model: type[BaseModel]) -> None:
        schema["description"] = describe_object(model.__doc__)

    model_config = ConfigDict(json_schema_extra=json_schema_extra)


class YamlConfigSettingsSource(PydanticBaseSettingsSource):
    """Loads variables from a YAML file at the project's root."""

    log = logging.getLogger()

    config_dir = Path()
    config_file_base_name = "config"
    environment: str | None = None

    def __init__(self, settings_cls) -> None:
        super().__init__(settings_cls)
        default_config = self.load_config(
            self.config_dir / f"{self.config_file_base_name}.yaml"
        )
        env_config = (
            self.load_config(
                self.config_dir
                / f"{self.config_file_base_name}_{self.environment}.yaml"
            )
            if self.environment
            else {}
        )
        self.config = update_nested_pair(env_config, default_config)

    @staticmethod
    def load_config(file: Path) -> dict:
        """Load YAML file if it exists.

        :param file: Path to a ``config*.yaml``
        :return: Dict containing the config or empty dict if file doesn't exist
        """
        if file.exists() and isinstance((loaded_file := load_yaml_file(file)), dict):
            return loaded_file
        return {}

    @override
    def get_field_value(
        self,
        field: FieldInfo,
        field_name: str,
    ) -> tuple[Any, str, bool]:
        return self.config.get(field_name), field_name, False

    @override
    def prepare_field_value(
        self, field_name: str, field: FieldInfo, value: Any, value_is_complex: bool
    ) -> Any:
        return value

    @override
    def __call__(self) -> dict[str, Any]:
        d: dict[str, Any] = {}
        for field_name, field in self.settings_cls.model_fields.items():
            field_value, field_key, value_is_complex = self.get_field_value(
                field,
                field_name,
            )
            field_value = self.prepare_field_value(
                field_name,
                field,
                field_value,
                value_is_complex,
            )
            if field_value is not None:
                d[field_key] = field_value
        return d



import logging
from functools import cached_property

from pydantic import Field, computed_field
from typing_extensions import override

from kpops.component_handlers.kubernetes.pvc_handler import PVCHandler
from kpops.components import HelmApp
from kpops.components.base_components.kafka_app import (
    KafkaApp,
    KafkaAppCleaner,
)
from kpops.components.base_components.models.topic import KafkaTopic
from kpops.components.streams_bootstrap import StreamsBootstrap
from kpops.components.streams_bootstrap.app_type import AppType
from kpops.components.streams_bootstrap.streams.model import (
    StreamsAppValues,
)
from kpops.utils.docstring import describe_attr

log = logging.getLogger("StreamsApp")


class StreamsAppCleaner(KafkaAppCleaner):
    from_: None = None
    to: None = None
    app: StreamsAppValues

    @property
    @override
    def helm_chart(self) -> str:
        return f"{self.repo_config.repository_name}/{AppType.CLEANUP_STREAMS_APP.value}"

    @override
    async def clean(self, dry_run: bool) -> None:
        await super().clean(dry_run)
        if self.app.stateful_set and self.app.persistence.enabled:
            await self.clean_pvcs(dry_run)

    async def clean_pvcs(self, dry_run: bool) -> None:
        app_full_name = super(HelmApp, self).full_name
        pvc_handler = await PVCHandler.create(app_full_name, self.namespace)
        if dry_run:
            pvc_names = await pvc_handler.list_pvcs()
            log.info(f"Deleting the PVCs {pvc_names} for StatefulSet '{app_full_name}'")
        else:
            log.info(f"Deleting the PVCs for StatefulSet '{app_full_name}'")
            await pvc_handler.delete_pvcs()


class StreamsApp(KafkaApp, StreamsBootstrap):
    """StreamsApp component that configures a streams-bootstrap app.

    :param app: Application-specific settings
    """

    app: StreamsAppValues = Field(
        default=...,
        description=describe_attr("app", __doc__),
    )

    @computed_field
    @cached_property
    def _cleaner(self) -> StreamsAppCleaner:
        return StreamsAppCleaner(
            config=self.config,
            handlers=self.handlers,
            **self.model_dump(by_alias=True, exclude={"_cleaner", "from_", "to"}),
        )

    @property
    @override
    def input_topics(self) -> list[KafkaTopic]:
        return self.app.streams.input_topics

    @property
    @override
    def extra_input_topics(self) -> dict[str, list[KafkaTopic]]:
        return self.app.streams.extra_input_topics

    @property
    @override
    def output_topic(self) -> KafkaTopic | None:
        return self.app.streams.output_topic

    @property
    @override
    def extra_output_topics(self) -> dict[str, KafkaTopic]:
        return self.app.streams.extra_output_topics

    @override
    def add_input_topics(self, topics: list[KafkaTopic]) -> None:
        self.app.streams.add_input_topics(topics)

    @override
    def add_extra_input_topics(self, role: str, topics: list[KafkaTopic]) -> None:
        self.app.streams.add_extra_input_topics(role, topics)

    @override
    def set_input_pattern(self, name: str) -> None:
        self.app.streams.input_pattern = name

    @override
    def add_extra_input_pattern(self, role: str, topic: str) -> None:
        self.app.streams.extra_input_patterns[role] = topic

    @override
    def set_output_topic(self, topic: KafkaTopic) -> None:
        self.app.streams.output_topic = topic

    @override
    def set_error_topic(self, topic: KafkaTopic) -> None:
        self.app.streams.error_topic = topic

    @override
    def add_extra_output_topic(self, topic: KafkaTopic, role: str) -> None:
        self.app.streams.extra_output_topics[role] = topic

    @property
    @override
    def helm_chart(self) -> str:
        return f"{self.repo_config.repository_name}/{AppType.STREAMS_APP.value}"

    @override
    async def reset(self, dry_run: bool) -> None:
        self._cleaner.app.streams.delete_output = False
        await self._cleaner.clean(dry_run)

    @override
    async def clean(self, dry_run: bool) -> None:
        self._cleaner.app.streams.delete_output = True
        await self._cleaner.clean(dry_run)



import logging

import httpx

log = logging.getLogger("HttpxException")


class HttpxException(Exception):
    def __init__(self, response: httpx.Response) -> None:
        self.error_code = response.status_code
        self.error_msg = "Something went wrong!"
        try:
            log.exception(
                f"The request responded with the code {self.error_code}. Error body: {response.json()}",
            )
            response.raise_for_status()
        except httpx.HTTPError as e:
            self.error_msg = str(e)
            log.exception(f"More information: {self.error_msg}")
        super().__init__()



import os
from collections import UserDict
from collections.abc import ItemsView, KeysView, MutableMapping, ValuesView

PIPELINE_PATH = "pipeline.path"


class Environment(UserDict[str, str]):
    """Internal environment wrapping OS environment."""

    def __init__(
        self, mapping: MutableMapping[str, str] | None = None, /, **kwargs: str
    ) -> None:
        self._global = os.environ
        if mapping is None:
            mapping = {}
        if kwargs:
            mapping.update(**kwargs)
        super().__init__(mapping)

    def __getitem__(self, key: str) -> str:
        try:
            return self.data[key]
        except KeyError:
            return self._global[key]

    def __contains__(self, key: object) -> bool:
        return super().__contains__(key) or self._global.__contains__(key)

    @property
    def _dict(self) -> dict[str, str]:
        return {**self._global, **self.data}

    def keys(self) -> KeysView[str]:
        return KeysView(self._dict)

    def values(self) -> ValuesView[str]:
        return ValuesView(self._dict)

    def items(self) -> ItemsView[str, str]:
        return ItemsView(self._dict)


ENV = Environment()



from __future__ import annotations

import logging

from kubernetes_asyncio import client, config
from kubernetes_asyncio.client import ApiClient

log = logging.getLogger("PVC_handler")


class PVCHandler:
    def __init__(self, app_name: str, namespace: str):
        self.app_name = app_name
        self.namespace = namespace

    @classmethod
    async def create(cls, app_name: str, namespace: str) -> PVCHandler:
        self = cls(app_name, namespace)
        await config.load_kube_config()
        return self

    async def list_pvcs(self) -> list[str]:
        async with ApiClient() as api:
            core_v1_api = client.CoreV1Api(api)
            pvc_list = await core_v1_api.list_namespaced_persistent_volume_claim(
                namespace=self.namespace, label_selector=f"app={self.app_name}"
            )

            pvc_names = [pvc.metadata.name for pvc in pvc_list.items]
            if not pvc_names:
                log.warning(
                    f"No PVCs found for app '{self.app_name}', in namespace '{self.namespace}'"
                )
            log.debug(
                f"In namespace '{self.namespace}' StatefulSet '{self.app_name}' has corresponding PVCs: '{pvc_names}'"
            )
            return pvc_names

    async def delete_pvcs(self) -> None:
        async with ApiClient() as api:
            core_v1_api = client.CoreV1Api(api)
            pvc_names = await self.list_pvcs()
            log.debug(
                f"Deleting in namespace '{self.namespace}' StatefulSet '{self.app_name}' PVCs '{pvc_names}'"
            )
            for pvc_name in pvc_names:
                await core_v1_api.delete_namespaced_persistent_volume_claim(
                    pvc_name, self.namespace
                )  # type: ignore [reportGeneralTypeIssues]



from functools import cached_property
from typing import Generic, TypeVar

_T = TypeVar("_T")
_R = TypeVar("_R")


class cached_classproperty(cached_property, Generic[_T, _R]):
    def __get__(self, owner_obj: _T, owner_cls: type[_T]) -> _R:
        return self.func(owner_cls)



import re
from collections import ChainMap as _ChainMap
from collections.abc import Mapping
from string import Template
from typing import Any, TypeVar

from typing_extensions import override

_V = TypeVar("_V", bound=object)


def update_nested_pair(
    original_dict: dict[str, _V], other_dict: Mapping[str, _V]
) -> dict[str, _V]:
    """Nested update for 2 dictionaries.

    Adds all new fields in ``other_dict`` to ``original_dict``.
    Does not update existing fields.
    To "update" a dict with new values in existing fields:
    ``original_dict={dict_with_new_values_and_fields}``
    ``other_dict={dict_to_be_updated}``

    :param original_dict: Dictionary to be updated
    :param other_dict: Mapping that contains new key-value pairs
    :return: Updated dictionary
    """
    for key, value in other_dict.items():
        if isinstance(value, Mapping):
            nested_val = original_dict.get(key, {})
            if isinstance(nested_val, dict):
                original_dict[key] = update_nested_pair(nested_val, value)
        elif key not in original_dict:
            original_dict[key] = value
    return original_dict


def update_nested(*argv: dict[str, _V]) -> dict[str, _V]:
    """Merge multiple configuration dicts.

    The dicts have multiple layers. These layers will be merged recursively.

    The leftmost arg has the highest priority, only new fields will be added to it.
    It "updates" the lower prio dict on the right.

    :param argv: n dictionaries
    :returns: Merged configuration dict
    """
    if len(argv) == 0:
        return {}
    if len(argv) == 1:
        return argv[0]
    if len(argv) == 2:
        return update_nested_pair(*argv)
    return update_nested(update_nested_pair(argv[0], argv[1]), *argv[2:])


def flatten_mapping(
    nested_mapping: Mapping[str, _V],
    prefix: str | None = None,
    separator: str = "_",
) -> dict[str, _V]:
    """Flattens a Mapping.

    :param nested_mapping: Nested mapping that is to be flattened
    :param prefix: Prefix that will be applied to all top-level keys in the output., defaults to None
    :param separator: Separator between the prefix and the keys, defaults to "_"
    :returns: "Flattened" mapping in the form of dict
    """
    if not isinstance(nested_mapping, Mapping):
        msg = "Argument nested_mapping is not a Mapping"
        raise TypeError(msg)
    top: dict[str, Any] = {}
    for key, value in nested_mapping.items():
        if not isinstance(key, str):
            msg = f"Argument nested_mapping contains a non-str key: {key}"
            raise TypeError(msg)
        if prefix:
            key = prefix + separator + key
        if isinstance(value, Mapping):
            nested_mapping = flatten_mapping(value, key, separator)
            top = update_nested_pair(top, nested_mapping)
        else:
            top[key] = value
    return top


def generate_substitution(
    input: dict[str, _V],
    prefix: str | None = None,
    existing_substitution: dict[str, _V] | None = None,
    separator: str | None = None,
) -> dict[str, _V]:
    """Generate a complete substitution dict from a given dict.

    Finds all attributes that belong to a model and expands them to create
    a dict containing each variable name and value to substitute with.

    :param input: Dict from which to generate the substitution
    :param prefix: Prefix the preceeds all substitution variables, defaults to None
    :param substitution: existing substitution to include
    :returns: Substitution dict of all variables related to the model.
    """
    existing_substitution = existing_substitution or {}
    if separator is None:
        return update_nested_pair(existing_substitution, flatten_mapping(input, prefix))
    return update_nested_pair(
        existing_substitution or {}, flatten_mapping(input, prefix, separator)
    )


_sentinel_dict = {}


class ImprovedTemplate(Template):
    """Introduces the dot as an allowed character in placeholders."""

    idpattern = r"(?a:[_a-z][_.a-z0-9]*)"

    @override
    def safe_substitute(self, mapping=_sentinel_dict, /, **kws) -> str:
        if mapping is _sentinel_dict:
            mapping = kws
        elif kws:
            mapping = _ChainMap(kws, mapping)

        # Helper function for .sub()
        def convert(mo: re.Match):
            named = mo.group("named") or mo.group("braced")
            if named is not None:
                try:
                    if "." not in named:
                        return str(mapping[named])
                    return str(mapping[named.replace(".", "__")])
                except KeyError:
                    return mo.group()
            if mo.group("escaped") is not None:
                return self.delimiter
            if mo.group("invalid") is not None:
                return mo.group()
            msg = "Unrecognized named group in pattern"
            raise ValueError(msg, self.pattern)

        return self.pattern.sub(convert, self.template)



import logging

log = logging.getLogger("docstring_utils")


def describe_attr(name: str, docstr: str | None) -> str:
    """Read attribute description from class docstring.

    **Works only with reStructuredText docstrings.**

    :param name: Attribute name
    :param docstr: Docstring from which to read. Note that the class' docstring is stored in attr ``__doc__``
    :returns: Description of the class attribute read from the class docstring
    """
    if docstr is None:
        return ""
    docstr = docstr.partition(f":param {name}:")[2]
    return _trim_description_end(docstr)


def describe_object(docstr: str | None) -> str:
    """Return description from an object's docstring.

    Excludes parameters and return definitions

    **Works only with reStructuredText docstrings.**

    :param docstr: The docstring
    :returns: Description taken from the docstring
    """
    if docstr is None:
        return ""

    # reST docstrings have a short description/title as their first line.
    # Optionally, they have a longer description below. Here we separate the
    # title from the rest with a newline as `_trim_description_end()`
    # removes all newlines.
    docstr = docstr.strip()
    if "\n" in docstr:
        title_end = docstr.index("\n")
        desc = docstr[:title_end] + "\n" + _trim_description_end(docstr[title_end:])
        return desc.rstrip()
    return _trim_description_end(docstr)


def _trim_description_end(desc: str) -> str:
    """Remove the unwanted text that comes after a description in a docstring.

    Also removes all whitespaces and newlines and replaces them with a single space.

    A description is defined here as a string of text written in natural language.
    A description ends at the occurence of a separator such as ``returns``.

    **Works only with reStructuredText docstrings.**

    :param desc: Description to be isolated, only the end will be trimmed
    :returns: Isolated description
    """
    desc_enders = [
        ":param ",
        ":returns:",
        ":raises:",
        "defaults to ",
    ]
    end_index = len(desc)
    for desc_ender in desc_enders:
        if (desc_ender in desc) and (desc.index(desc_ender) < end_index):
            end_index = desc.index(desc_ender)
    desc_split = desc[:end_index].split()
    for line in desc_split:
        line = line.rstrip()
    desc = " ".join(desc_split)
    return desc.rstrip(",").rstrip()



from __future__ import annotations

import logging
from abc import ABC
from collections.abc import Callable
from typing import Any

import pydantic
from pydantic import AliasChoices, ConfigDict, Field
from typing_extensions import override

from kpops.components.base_components.cleaner import Cleaner
from kpops.components.base_components.helm_app import HelmAppValues
from kpops.components.base_components.models.topic import KafkaTopic, KafkaTopicStr
from kpops.components.base_components.pipeline_component import PipelineComponent
from kpops.components.streams_bootstrap import StreamsBootstrap
from kpops.utils.docstring import describe_attr
from kpops.utils.pydantic import (
    CamelCaseConfigModel,
    DescConfigModel,
    exclude_by_value,
    exclude_defaults,
)

log = logging.getLogger("KafkaApp")


class KafkaStreamsConfig(CamelCaseConfigModel, DescConfigModel):
    """Kafka Streams config.

    :param brokers: Brokers
    :param schema_registry_url: URL of the schema registry, defaults to None
    :param extra_output_topics: Extra output topics
    :param output_topic: Output topic, defaults to None
    """

    brokers: str = Field(default=..., description=describe_attr("brokers", __doc__))
    schema_registry_url: str | None = Field(
        default=None,
        validation_alias=AliasChoices(
            "schema_registry_url", "schemaRegistryUrl"
        ),  # TODO: same for other camelcase fields, avoids duplicates during enrichment
        description=describe_attr("schema_registry_url", __doc__),
    )
    extra_output_topics: dict[str, KafkaTopicStr] = Field(
        default={}, description=describe_attr("extra_output_topics", __doc__)
    )
    output_topic: KafkaTopicStr | None = Field(
        default=None,
        description=describe_attr("output_topic", __doc__),
        json_schema_extra={},
    )

    model_config = ConfigDict(extra="allow")

    @pydantic.field_validator("extra_output_topics", mode="before")
    @classmethod
    def deserialize_extra_output_topics(
        cls, extra_output_topics: Any
    ) -> dict[str, KafkaTopic] | Any:
        if isinstance(extra_output_topics, dict):
            return {
                role: KafkaTopic(name=topic_name)
                for role, topic_name in extra_output_topics.items()
            }
        return extra_output_topics

    @pydantic.field_serializer("extra_output_topics")
    def serialize_extra_output_topics(
        self, extra_topics: dict[str, KafkaTopic]
    ) -> dict[str, str]:
        return {role: topic.name for role, topic in extra_topics.items()}

    # TODO(Ivan Yordanov): Currently hacky and potentially unsafe. Find cleaner solution
    @pydantic.model_serializer(mode="wrap", when_used="always")
    def serialize_model(
        self, handler: Callable, info: pydantic.SerializationInfo
    ) -> dict[str, Any]:
        return exclude_defaults(self, exclude_by_value(handler(self), None))


class KafkaAppValues(HelmAppValues):
    """Settings specific to Kafka Apps.

    :param streams: Kafka streams config
    """

    streams: KafkaStreamsConfig = Field(
        default=..., description=describe_attr("streams", __doc__)
    )


class KafkaAppCleaner(Cleaner, StreamsBootstrap, ABC):
    """Helm app for resetting and cleaning a streams-bootstrap app."""

    from_: None = None
    to: None = None

    @property
    @override
    def helm_chart(self) -> str:
        raise NotImplementedError

    @override
    async def clean(self, dry_run: bool) -> None:
        """Clean an app using a cleanup job.

        :param dry_run: Dry run command
        """
        log.info(f"Uninstall old cleanup job for {self.helm_release_name}")
        await self.destroy(dry_run)

        log.info(f"Init cleanup job for {self.helm_release_name}")
        await self.deploy(dry_run)

        if not self.config.retain_clean_jobs:
            log.info(f"Uninstall cleanup job for {self.helm_release_name}")
            await self.destroy(dry_run)


class KafkaApp(PipelineComponent, ABC):
    """Base component for Kafka-based components.

    Producer or streaming apps should inherit from this class.

    :param app: Application-specific settings
    """

    app: KafkaAppValues = Field(
        default=...,
        description=describe_attr("app", __doc__),
    )

    @override
    async def deploy(self, dry_run: bool) -> None:
        if self.to:
            for topic in self.to.kafka_topics:
                await self.handlers.topic_handler.create_topic(topic, dry_run=dry_run)

            if self.handlers.schema_handler:
                await self.handlers.schema_handler.submit_schemas(
                    to_section=self.to, dry_run=dry_run
                )

        await super().deploy(dry_run)



import json
from typing import Any


def is_jsonable(input: Any) -> bool:
    """Check whether a value is json-serializable.

    :param input: Value to be checked.
    """
    try:
        json.dumps(input)
    except (TypeError, OverflowError):
        return False
    else:
        return True



from kpops.component_handlers.utils.exception import HttpxException


class TopicNotFoundException(Exception):
    pass


class TopicTransactionError(Exception):
    pass


class KafkaRestProxyError(HttpxException):
    pass



__version__ = "5.1.1"

# export public API functions
from kpops.cli.main import clean, deploy, destroy, generate, init, manifest, reset

__all__ = (
    "generate",
    "manifest",
    "deploy",
    "destroy",
    "reset",
    "clean",
    "init",
)



from kpops.components.base_components.helm_app import HelmApp
from kpops.components.base_components.kafka_app import KafkaApp
from kpops.components.base_components.kafka_connector import (
    KafkaConnector,
    KafkaSinkConnector,
    KafkaSourceConnector,
)
from kpops.components.base_components.kubernetes_app import KubernetesApp
from kpops.components.base_components.pipeline_component import PipelineComponent
from kpops.components.streams_bootstrap import StreamsBootstrap
from kpops.components.streams_bootstrap.producer.producer_app import ProducerApp
from kpops.components.streams_bootstrap.streams.streams_app import StreamsApp

__all__ = (
    "HelmApp",
    "KafkaApp",
    "KafkaConnector",
    "KafkaSinkConnector",
    "KafkaSourceConnector",
    "KubernetesApp",
    "StreamsBootstrap",
    "ProducerApp",
    "StreamsApp",
    "PipelineComponent",
    "StreamsApp",
    "ProducerApp",
)



from abc import ABC

from typing_extensions import override

from kpops.component_handlers.helm_wrapper.model import HelmFlags
from kpops.component_handlers.helm_wrapper.utils import (
    create_helm_name_override,
    create_helm_release_name,
)
from kpops.components.base_components.helm_app import HelmApp


class Cleaner(HelmApp, ABC):
    """Generic Helm app for cleaning or resetting."""

    suffix: str = "-clean"

    @property
    @override
    def full_name(self) -> str:
        return super().full_name + self.suffix

    @property
    @override
    def helm_release_name(self) -> str:
        return create_helm_release_name(self.full_name, self.suffix)

    @property
    @override
    def helm_name_override(self) -> str:
        return create_helm_name_override(self.full_name, self.suffix)

    @property
    @override
    def helm_flags(self) -> HelmFlags:
        return HelmFlags(
            create_namespace=self.config.create_namespace,
            version=self.version,
            wait=True,
            wait_for_jobs=True,
        )
